__all__ = ("BaseObs",)

import os
import warnings

import numpy as np
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import LsstCameraFootprint, angular_separation

from rubin_sim.phot_utils import Bandpass, Sed

from .ooephemerides import PyOrbEphemerides


class BaseObs:
    """
    Base class to generate observations of a set of moving objects.

    Parameters
    ----------
    footPrint: `str`, optional
        Specify the footprint for the FOV. Options include "camera", "circle", "rectangle".
        'Camera' means use the actual LSST camera footprint (following a rough cut with a circular FOV).
        Default is camera FOV.
    r_fov : `float`, optional
        If footprint is "circular", this is the radius of the fov (in degrees).
        Default 1.75 degrees (only used for circular fov).
    x_tol : `float`, optional
        If footprint is "rectangular", this is half of the width of the (on-sky) fov in the RA
        direction (in degrees).
        Default 5 degrees. (so size of footprint in degrees will be 10 degrees in the RA direction).
    y_tol : `float`, optional
        If footprint is "rectangular", this is half of the width of the fov in Declination (in degrees).
        Default is 3 degrees (so size of footprint in degrees will be 6 degrees in the Dec direction).
    eph_mode: `str`, optional
        Mode for ephemeris generation - nbody or 2body. Default is nbody.
    eph_type: `str`, optional
        Type of ephemerides to generate - full or basic.
        Full includes all values calculated by openorb; Basic includes a more basic set.
        Default is Basic.  (this includes enough information for most standard MAF metrics).
    eph_file: `str` or None, optional
        The name of the planetary ephemerides file to use for ephemeris generation.
        Default (None) will use the default for PyOrbEphemerides.
    obs_code: `str`, optional
        Observatory code for ephemeris generation. Default is "I11" - Cerro Pachon.
    obs_time_col: `str`, optional
        Name of the time column in the obsData. Default 'observationStartMJD'.
    obs_time_scale: `str`, optional
        Type of timescale for MJD (TAI or UTC currently). Default TAI.
    seeing_col: `str`, optional
        Name of the seeing column in the obsData. Default 'seeingFwhmGeom'.
        This should be the geometric/physical seeing as it is used for the trailing loss calculation.
    visit_exp_time_col: `str`, optional
        Name of the visit exposure time column in the obsData. Default 'visitExposureTime'.
    obs_ra: `str`, optional
        Name of the RA column in the obsData. Default 'fieldRA'.
    obs_dec: `str`, optional
        Name of the Dec column in the obsData. Default 'fieldDec'.
    obs_rot_sky_pos: `str`, optional
        Name of the Rotator column in the obsData. Default 'rotSkyPos'.
    obs_degrees: `bool`, optional
        Whether the observational data is in degrees or radians. Default True (degrees).
    outfile_name : `str`, optional
        The output file name.
        Default is 'lsst_obs.dat'.
    obs_info : `str`, optional
        A string that captures provenance information about the observations.
        For example: 'baseline_v2.0_10yrs, years 0-5' or 'baseline2018a minus NES'
        Default ''.
    """

    def __init__(
        self,
        footprint="camera",
        r_fov=1.75,
        x_tol=5,
        y_tol=3,
        eph_mode="nbody",
        eph_type="basic",
        obs_code="I11",
        eph_file=None,
        obs_time_col="observationStartMJD",
        obs_time_scale="TAI",
        seeing_col="seeingFwhmGeom",
        visit_exp_time_col="visitExposureTime",
        obs_ra="fieldRA",
        obs_dec="fieldDec",
        obs_rot_sky_pos="rotSkyPos",
        obs_degrees=True,
        outfile_name="lsst_obs.dat",
        obs_info="",
        camera_footprint_file=None,
    ):
        # Strings relating to the names of columns in the observation metadata.
        self.obs_code = obs_code
        self.obs_time_col = obs_time_col
        self.obs_time_scale = obs_time_scale
        self.seeing_col = seeing_col
        self.visit_exp_time_col = visit_exp_time_col
        self.obs_ra = obs_ra
        self.obs_dec = obs_dec
        self.obs_rot_sky_pos = obs_rot_sky_pos
        self.obs_degrees = obs_degrees
        # Save a space for the standard object colors.
        self.colors = {}
        self.outfile_name = outfile_name
        # Values for identifying observations.
        self.footprint = footprint.lower()
        if self.footprint == "camera":
            self._setup_camera(camera_footprint_file=camera_footprint_file)
        self.r_fov = r_fov
        self.x_tol = x_tol
        self.y_tol = y_tol
        # Values for ephemeris generation.
        if eph_mode.lower() not in ("2body", "nbody"):
            raise ValueError("Ephemeris generation must be 2body or nbody.")
        self.eph_mode = eph_mode
        self.eph_type = eph_type
        self.eph_file = eph_file

        self._setup_info(obs_info=obs_info)

    def _setup_info(self, obs_info=""):
        """Generate a dict to record relevant settings"""
        info = {
            "obs_info": obs_info,
            "footprint": self.footprint,
            "eph_mode": self.eph_mode,
            "eph_type": self.eph_type,
        }
        if self.footprint == "circle":
            info["r_fov"] = self.r_fov
        if self.footprint == "rectangle":
            info["xtol"] = self.x_tol
            info["ytol"] = self.y_tol

        # convert to numpy array in anticipation of saving
        names = list(info.keys())
        types = [np.array(info[key]).dtype for key in names]
        self.info = np.zeros(1, dtype=list(zip(names, types)))
        for key in names:
            self.info[key] = info[key]

    def _setup_camera(self, camera_footprint_file=None):
        self.camera = LsstCameraFootprint(units="degrees", footprint_file=camera_footprint_file)

    def setup_ephemerides(self):
        """Initialize the ephemeris generator. Save the setup PyOrbEphemeris class.

        This uses the default engine, pyoorb - however this could be overwritten to use another generator.
        """
        self.ephems = PyOrbEphemerides(ephfile=self.eph_file)

    def generate_ephemerides(self, sso, times, eph_mode=None, eph_type=None):
        """Generate ephemerides for 'sso' at times 'times' (assuming MJDs, with timescale self.obs_time_scale).

        The default engine here is pyoorb, however this method could be overwritten to use another ephemeris
        generator, such as ADAM.

        The initialized pyoorb class (PyOrbEphemerides) is saved, to skip setup on subsequent calls.

        Parameters
        ----------
        sso : `rubin_sim.movingObjects.Orbits`
            Typically this will be a single object.
        times: `np.ndarray`
            The times at which to generate ephemerides. MJD.
        eph_mode: `str` or None, optional
            Potentially override default eph_mode (self.eph_mode). Must be '2body' or 'nbody'.

        Returns
        -------
        ephs : `pd.Dataframe`
            Results from propigating the orbit(s) to the specified times. Columns like:
            obj_id, sedname, time, ra, dec, dradt, ddecdt, phase, solarelon, etc
        """
        if not hasattr(self, "ephems"):
            self.setup_ephemerides()
        if eph_mode is None:
            eph_mode = self.eph_mode
        if eph_type is None:
            eph_type = self.eph_type
        self.ephems.set_orbits(sso)
        ephs = self.ephems.generate_ephemerides(
            times,
            time_scale=self.obs_time_scale,
            obscode=self.obs_code,
            eph_mode=eph_mode,
            eph_type=eph_type,
            by_object=True,
        )
        return ephs

    def calc_trailing_losses(self, velocity, seeing, texp=30.0):
        """Calculate the detection and SNR trailing losses.

        'Trailing' losses = loss in sensitivity due to the photons from the source being
        spread over more pixels; thus more sky background is included when calculating the
        flux from the object and thus the SNR is lower than for an equivalent brightness
        stationary/PSF-like source. dmagTrail represents this loss.

        'Detection' trailing losses = loss in sensitivity due to the photons from the source being
        spread over more pixels, in a non-stellar-PSF way, while source detection is (typically) done
        using a stellar PSF filter and 5-sigma cutoff values based on assuming peaks from stellar PSF's
        above the background; thus the SNR is lower than for an equivalent brightness stationary/PSF-like
        source (and by a greater factor than just the simple SNR trailing loss above).
        dmag_detect represents this loss.

        Parameters
        ----------
        velocity : `np.ndarray` or `float`
            The velocity of the moving objects, in deg/day.
        seeing : `np.ndarray` or `float`
            The seeing of the images, in arcseconds.
        texp : `np.ndarray` or `float`, optional
            The exposure time of the images, in seconds. Default 30.

        Returns
        -------
        dmag Trail, dmag_detect : (`np.ndarray`, `np.ndarray`) or (`float`, `float`)
            dmag_trail and dmag_detect for each set of velocity/seeing/texp values.
        """
        a_trail = 0.761
        b_trail = 1.162
        a_det = 0.420
        b_det = 0.003
        x = velocity * texp / seeing / 24.0
        dmag_trail = 1.25 * np.log10(1 + a_trail * x**2 / (1 + b_trail * x))
        dmag_detect = 1.25 * np.log10(1 + a_det * x**2 / (1 + b_det * x))
        return (dmag_trail, dmag_detect)

    def read_filters(
        self,
        filter_dir=None,
        bandpass_root="total_",
        bandpass_suffix=".dat",
        filterlist=("u", "g", "r", "i", "z", "y"),
        v_dir=None,
        v_filter="harris_V.dat",
    ):
        """
        Read (LSST) and Harris (V) filter throughput curves.

        Only the defaults are LSST specific; this can easily be adapted for any survey.

        Parameters
        ----------
        filter_dir : `str`, optional
            Directory containing the filter throughput curves ('total*.dat')
            Default set by 'LSST_THROUGHPUTS_BASELINE' env variable.
        bandpass_root : `str`, optional
            Rootname of the throughput curves in filterlist.
            E.g. throughput curve names are bandpass_root + filterlist[i] + bandpass_suffix
            Default total\_ (appropriate for LSST throughput repo).
        bandpass_suffix : `str`, optional
            Suffix for the throughput curves in filterlist.
            Default '.dat' (appropriate for LSST throughput repo).
        filterlist : `list`, optional
            List containing the filter names to use to calculate colors.
            Default ('u', 'g', 'r', 'i', 'z', 'y')
        v_dir : `str`, optional
            Directory containing the V band throughput curve.
            Default None = $SIMS_MOVINGOBJECTS_DIR/data.
        v_filter : `str`, optional
            Name of the V band filter curve.
            Default harris_V.dat.
        """
        if filter_dir is None:
            filter_dir = os.path.join(get_data_dir(), "throughputs/baseline")
        if v_dir is None:
            v_dir = os.path.join(get_data_dir(), "movingObjects")
        self.filterlist = filterlist
        # Read filter throughput curves from disk.
        self.lsst = {}
        for f in self.filterlist:
            self.lsst[f] = Bandpass()
            self.lsst[f].read_throughput(os.path.join(filter_dir, bandpass_root + f + bandpass_suffix))
        self.vband = Bandpass()
        self.vband.read_throughput(os.path.join(v_dir, v_filter))

    def calc_colors(self, sedname="C.dat", sed_dir=None):
        """Calculate the colors for a given SED.

        If the sedname is not already in the dictionary self.colors, this reads the
        SED from disk and calculates all V-[filter] colors for all filters in self.filterlist.
        The result is stored in self.colors[sedname][filter], so will not be recalculated if
        the SED + color is reused for another object.

        Parameters
        ----------
        sedname : `str`, optional
            Name of the SED. Default 'C.dat'.
        sed_dir : `str`, optional
            Directory containing the SEDs of the moving objects.
            Default None = $SIMS_MOVINGOBJECTS_DIR/data.

        Returns
        -------
        colors : `dict`
            Dictionary of the colors in self.filterlist for this particular Sed.
        """
        if sedname not in self.colors:
            if sed_dir is None:
                sed_dir = os.path.join(get_data_dir(), "movingObjects")
            mo_sed = Sed()
            mo_sed.read_sed_flambda(os.path.join(sed_dir, sedname))
            vmag = mo_sed.calc_mag(self.vband)
            self.colors[sedname] = {}
            for f in self.filterlist:
                self.colors[sedname][f] = mo_sed.calc_mag(self.lsst[f]) - vmag
        return self.colors[sedname]

    def sso_in_circle_fov(self, ephems, obs_data):
        """Determine which observations are within a circular fov for a series of observations.
        Note that ephems and obs_data must be the same length.

        Parameters
        ----------
        ephems : `np.recarray`
            Ephemerides for the objects.
        obs_data : `np.recarray`
            The observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object observations which are inside the fov.
        """
        return self._sso_in_circle_fov(ephems, obs_data, self.r_fov)

    def _sso_in_circle_fov(self, ephems, obs_data, r_fov):
        if not self.obs_degrees:
            sep = angular_separation(
                ephems["ra"],
                ephems["dec"],
                np.degrees(obs_data[self.obs_ra]),
                np.degrees(obs_data[self.obs_dec]),
            )
        else:
            sep = angular_separation(
                ephems["ra"],
                ephems["dec"],
                obs_data[self.obs_ra],
                obs_data[self.obs_dec],
            )
        idx_obs = np.where(sep <= r_fov)[0]
        return idx_obs

    def sso_in_rectangle_fov(self, ephems, obs_data):
        """Determine which observations are within a rectangular FoV for a series of observations.
        Note that ephems and obs_data must be the same length.

        Parameters
        ----------
        ephems : `np.recarray`
            Ephemerides for the objects.
        obs_data : `np.recarray`
            The observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object observations which are inside the fov.
        """
        return self._sso_in_rectangle_fov(ephems, obs_data, self.x_tol, self.y_tol)

    def _sso_in_rectangle_fov(self, ephems, obs_data, x_tol, y_tol):
        delta_dec = np.abs(ephems["dec"] - obs_data[self.obs_dec])
        delta_ra = np.abs((ephems["ra"] - obs_data[self.obs_ra]) * np.cos(np.radians(obs_data[self.obs_dec])))
        idx_obs = np.where((delta_dec <= y_tol) & (delta_ra <= x_tol))[0]
        return idx_obs

    def sso_in_camera_fov(self, ephems, obs_data):
        """Determine which observations are within the actual camera footprint for a series of observations.
        Note that ephems and obs_data must be the same length.

        Parameters
        ----------
        ephems : `np.ndarray`
            Ephemerides for the objects.
        obs_data : `np.ndarray`
            Observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object observations which are inside the fov.
        """
        if not hasattr(self, "camera"):
            self._setup_camera()

        if not self.obs_degrees:
            idx = self.camera(
                ephems["ra"],
                ephems["dec"],
                np.degrees(obs_data[self.obs_ra]),
                np.degrees(obs_data[self.obs_dec]),
                np.degrees(obs_data[self.obs_rot_sky_pos]),
            )
        else:
            idx = self.camera(
                ephems["ra"],
                ephems["dec"],
                obs_data[self.obs_ra],
                obs_data[self.obs_dec],
                obs_data[self.obs_rot_sky_pos],
            )
        return idx

    def sso_in_fov(self, ephems, obs_data):
        """Convenience layer - determine which footprint method to apply (from self.footprint) and use it.

        Parameters
        ----------
        ephems : `np.ndarray`
            Ephemerides for the objects.
        obs_data : `np.ndarray`
            Observation pointings.

        Returns
        -------
        indices : `np.ndarray`
            Returns the indexes of the numpy array of the object observations which are inside the fov.
        """
        if self.footprint == "camera":
            return self.sso_in_camera_fov(ephems, obs_data)
        elif self.footprint == "rectangle":
            return self.sso_in_rectangle_fov(ephems, obs_data)
        elif self.footprint == "circle":
            return self.sso_in_circle_fov(ephems, obs_data)
        else:
            warnings.warn("Using circular fov; could not match specified footprint.")
            self.footprint = "circle"
            return self.sso_in_circle_fov(ephems, obs_data)



__all__ = ("TransientMetric",)


import numpy as np

from .base_metric import BaseMetric


class TransientMetric(BaseMetric):
    """
    Calculate what fraction of the transients would be detected. Best paired with a spatial slicer.
    We are assuming simple light curves with no color evolution.

    Parameters
    ----------
    trans_duration : float, optional
        How long the transient lasts (days). Default 10.
    peak_time : float, optional
        How long it takes to reach the peak magnitude (days). Default 5.
    rise_slope : float, optional
        Slope of the light curve before peak time (mags/day).
        This should be negative since mags are backwards (magnitudes decrease towards brighter fluxes).
        Default 0.
    decline_slope : float, optional
        Slope of the light curve after peak time (mags/day).
        This should be positive since mags are backwards. Default 0.
    uPeak : float, optional
        Peak magnitude in u band. Default 20.
    gPeak : float, optional
        Peak magnitude in g band. Default 20.
    rPeak : float, optional
        Peak magnitude in r band. Default 20.
    iPeak : float, optional
        Peak magnitude in i band. Default 20.
    zPeak : float, optional
        Peak magnitude in z band. Default 20.
    yPeak : float, optional
        Peak magnitude in y band. Default 20.
    survey_duration : float, optional
        Length of survey (years).
        Default 10.
    survey_start : float, optional
        MJD for the survey start date.
        Default None (uses the time of the first observation).
    detect_m5_plus : float, optional
        An observation will be used if the light curve magnitude is brighter than m5+detect_m5_plus.
        Default 0.
    n_pre_peak : int, optional
        Number of observations (in any filter(s)) to demand before peak_time,
        before saying a transient has been detected.
        Default 0.
    n_per_lc : int, optional
        Number of sections of the light curve that must be sampled above the detect_m5_plus theshold
        (in a single filter) for the light curve to be counted.
        For example, setting n_per_lc = 2 means a light curve  is only considered detected if there
        is at least 1 observation in the first half of the LC, and at least one in the second half of the LC.
        n_per_lc = 4 means each quarter of the light curve must be detected to count.
        Default 1.
    n_filters : int, optional
        Number of filters that need to be observed for an object to be counted as detected.
        Default 1.
    n_phase_check : int, optional
        Sets the number of phases that should be checked.
        One can imagine pathological cadences where many objects pass the detection criteria,
        but would not if the observations were offset by a phase-shift.
        Default 1.
    count_method : {'full' 'partialLC'}, defaults to 'full'
        Sets the method of counting max number of transients. if 'full', the
        only full light curves that fit the survey duration are counted. If
        'partialLC', then the max number of possible transients is taken to be
        the integer floor
    """

    def __init__(
        self,
        metric_name="TransientDetectMetric",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        trans_duration=10.0,
        peak_time=5.0,
        rise_slope=0.0,
        decline_slope=0.0,
        survey_duration=10.0,
        survey_start=None,
        detect_m5_plus=0.0,
        u_peak=20,
        g_peak=20,
        r_peak=20,
        i_peak=20,
        z_peak=20,
        y_peak=20,
        n_pre_peak=0,
        n_per_lc=1,
        n_filters=1,
        n_phase_check=1,
        count_method="full",
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        super(TransientMetric, self).__init__(
            col=[self.mjd_col, self.m5_col, self.filter_col],
            units="Fraction Detected",
            metric_name=metric_name,
            **kwargs,
        )
        self.peaks = {
            "u": u_peak,
            "g": g_peak,
            "r": r_peak,
            "i": i_peak,
            "z": z_peak,
            "y": y_peak,
        }
        self.trans_duration = trans_duration
        self.peak_time = peak_time
        self.rise_slope = rise_slope
        self.decline_slope = decline_slope
        self.survey_duration = survey_duration
        self.survey_start = survey_start
        self.detect_m5_plus = detect_m5_plus
        self.n_pre_peak = n_pre_peak
        self.n_per_lc = n_per_lc
        self.n_filters = n_filters
        self.n_phase_check = n_phase_check
        self.count_method = count_method

    def light_curve(self, time, filters):
        """
        Calculate the magnitude of the object at each time, in each filter.

        Parameters
        ----------
        time : numpy.ndarray
            The times of the observations.
        filters : numpy.ndarray
            The filters of the observations.

        Returns
        -------
        numpy.ndarray
            The magnitudes of the object at each time, in each filter.
        """
        lc_mags = np.zeros(time.size, dtype=float)
        rise = np.where(time <= self.peak_time)
        lc_mags[rise] += self.rise_slope * time[rise] - self.rise_slope * self.peak_time
        decline = np.where(time > self.peak_time)
        lc_mags[decline] += self.decline_slope * (time[decline] - self.peak_time)
        for key in self.peaks:
            f_match = np.where(filters == key)
            lc_mags[f_match] += self.peaks[key]
        return lc_mags

    def run(self, data_slice, slice_point=None):
        """ "
        Calculate the detectability of a transient with the specified lightcurve.

        Parameters
        ----------
        data_slice : numpy.array
            Numpy structured array containing the data related to the visits provided by the slicer.
        slice_point : dict, optional
            Dictionary containing information about the slice_point currently active in the slicer.

        Returns
        -------
        float
            The total number of transients that could be detected.
        """
        # Total number of transients that could go off back-to-back
        if self.count_method == "partialLC":
            _n_trans_max = np.ceil(self.survey_duration / (self.trans_duration / 365.25))
        else:
            _n_trans_max = np.floor(self.survey_duration / (self.trans_duration / 365.25))
        tshifts = np.arange(self.n_phase_check) * self.trans_duration / float(self.n_phase_check)
        n_detected = 0
        n_trans_max = 0
        for tshift in tshifts:
            # Compute the total number of back-to-back transients are possible to detect
            # given the survey duration and the transient duration.
            n_trans_max += _n_trans_max
            if tshift != 0:
                n_trans_max -= 1
            if self.survey_start is None:
                survey_start = data_slice[self.mjd_col].min()
            time = (data_slice[self.mjd_col] - survey_start + tshift) % self.trans_duration

            # Which lightcurve does each point belong to
            lc_number = np.floor((data_slice[self.mjd_col] - survey_start) / self.trans_duration)

            lc_mags = self.light_curve(time, data_slice[self.filter_col])

            # How many criteria needs to be passed
            detect_thresh = 0

            # Flag points that are above the SNR limit
            detected = np.zeros(data_slice.size, dtype=int)
            detected[np.where(lc_mags < data_slice[self.m5_col] + self.detect_m5_plus)] += 1
            detect_thresh += 1

            # If we demand points on the rise
            if self.n_pre_peak > 0:
                detect_thresh += 1
                ord = np.argsort(data_slice[self.mjd_col])
                data_slice = data_slice[ord]
                detected = detected[ord]
                lc_number = lc_number[ord]
                time = time[ord]
                ulc_number = np.unique(lc_number)
                left = np.searchsorted(lc_number, ulc_number)
                right = np.searchsorted(lc_number, ulc_number, side="right")
                # Note here I'm using np.searchsorted to basically do a 'group by'
                # might be clearer to use scipy.ndimage.measurements.find_objects or pandas, but
                # this numpy function is known for being efficient.
                for le, ri in zip(left, right):
                    # Number of points where there are a detection
                    good = np.where(time[le:ri] < self.peak_time)
                    nd = np.sum(detected[le:ri][good])
                    if nd >= self.n_pre_peak:
                        detected[le:ri] += 1

            # Check if we need multiple points per light curve or multiple filters
            if (self.n_per_lc > 1) | (self.n_filters > 1):
                # make sure things are sorted by time
                ord = np.argsort(data_slice[self.mjd_col])
                data_slice = data_slice[ord]
                detected = detected[ord]
                lc_number = lc_number[ord]
                time = time[ord]
                ulc_number = np.unique(lc_number)
                left = np.searchsorted(lc_number, ulc_number)
                right = np.searchsorted(lc_number, ulc_number, side="right")
                detect_thresh += self.n_filters

                for le, ri in zip(left, right):
                    points = np.where(detected[le:ri] > 0)
                    ufilters = np.unique(data_slice[self.filter_col][le:ri][points])
                    phase_sections = np.floor(time[le:ri][points] / self.trans_duration * self.n_per_lc)
                    for filt_name in ufilters:
                        good = np.where(data_slice[self.filter_col][le:ri][points] == filt_name)
                        if np.size(np.unique(phase_sections[good])) >= self.n_per_lc:
                            detected[le:ri] += 1

            # Find the unique number of light curves that passed the required number of conditions
            n_detected += np.size(np.unique(lc_number[np.where(detected >= detect_thresh)]))

        # Rather than keeping a single "detected" variable, maybe make a mask for each criteria, then
        # reduce functions like: reduce_singleDetect, reduce_NDetect, reduce_PerLC, reduce_perFilter.
        # The way I'm running now it would speed things up.

        return float(n_detected) / n_trans_max



"""Summary metric plotting functions.
"""

__all__ = (
    "normalize_metric_summaries",
    "plot_run_metric",
    "plot_run_metric_mesh",
    "find_family_lines",
)

# imports
import warnings

import colorcet
import cycler
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

# constants

RUN_COL_NAMES = ["run", "OpsimRun", "run_name"]
METRIC_COL_NAMES = ["metric"]

# exception classes

# interface functions


def normalize_metric_summaries(
    baseline_run,
    summary,
    metric_sets=None,
):
    """Create a normalized `pandas.DataFrame` of metric summary values.

    Parameters
    ----------
    baseline_run : `str` or `list` of `str
        The name of the run that defines a normalized value of 1.
        If a list is provided, the median value of each metric across that list is used as the reference.
    summary : `pandas.DataFrame`
        The summary metrics to normalize (as returned by `get_metric_summaries`)
    metric_sets : `pandas.DataFrame`
        Metric metadata as returned by `archive.get_metric_sets`

    Returns
    -------
    norm_summary : `pandas.DataFrame`
        Metric summary values are returned in a `pandas.DataFrame`, with each
        column providing the metrics for one run, and each row the values for
        one metric. The metric names constitute the columns, and the
        index names are the canonical run names.
        Values of 1 indicate metric values that match that of the baseline,
        differences with with indicate fractional improvement (if > 1) or
        degradation (if < 1).
    """

    # If the DataFrame is transposed but the axes are properly
    # labeled, just fix it.
    cols_are_runs = summary.columns.name in RUN_COL_NAMES
    rows_are_metrics = summary.index.name in METRIC_COL_NAMES
    if cols_are_runs or rows_are_metrics:
        summary = summary.T

    runs = summary.index.values

    # Use only those metrics present both in the
    # summary and metrics sets dataframe
    if metric_sets is None:
        summary = summary.copy()
        used_metrics = summary.columns.values
    else:
        used_metrics = [s for s in summary.columns.values if s in metric_sets.metric.values]
        summary = summary.loc[:, used_metrics].copy()

    if summary.columns.name is None:
        summary.rename_axis(columns="metric", inplace=True)

    if summary.index.name in [None] + RUN_COL_NAMES:
        summary.rename_axis("run", inplace=True)

    # Get rid of duplicate metrics and runs
    summary = summary.T.groupby("metric").first().T.groupby("run").first()

    # And now create a line just for "baseline" --
    # if baseline_run is >1, this is created from the median values per metric of those runs
    # Make up a nonsense name for the reference, that is not currently in the summary dataframe
    baseline_comparison = "bbb"
    while baseline_comparison in summary.index:
        baseline_comparison += "b"

    if isinstance(summary.loc[baseline_run], pd.DataFrame):
        summary.loc[baseline_comparison] = summary.loc[baseline_run].median(axis="rows")
    else:
        summary.loc[baseline_comparison] = summary.loc[baseline_run]

    if metric_sets is None:
        # If no invert/mag - just do simple normalization (1 + (x-0)/x0)
        norm_summary = 1 + (summary.loc[:, :].sub(summary.loc[baseline_comparison, :], axis="columns")).div(
            summary.loc[baseline_comparison, :], axis="columns"
        )
    else:
        # Reindex metric set and remove duplicates or non-available metrics
        metric_names = [n for n in metric_sets.index.names if not n == "metric"]
        metric_sets = (
            metric_sets.reset_index(metric_names).groupby(level="metric").first().loc[used_metrics, :].copy()
        )

        norm_summary = pd.DataFrame(
            columns=summary.columns,
            index=summary.index,
            dtype="float",
        )

        # Direct metrics are those that are neither inverted, nor compared as magnitudes
        # direct = 1 + (value - norm) / norm == value / norm
        direct = ~np.logical_or(metric_sets["invert"], metric_sets["mag"])
        norm_summary.loc[:, direct] = summary.loc[:, direct]

        # invert = 1 + (1/value - 1/norm) / (1/norm) == norm / value
        norm_summary.loc[:, metric_sets["invert"]] = 1.0 / summary.loc[:, metric_sets["invert"]]

        # mag = 1 + (1+value-norm - (1+norm-norm)) / (1+norm-norm) == 1 + (value - norm)
        norm_summary.loc[:, metric_sets["mag"]] = 1.0 + summary.loc[
            :,
            metric_sets["mag"],
        ].subtract(summary.loc[baseline_comparison, metric_sets["mag"]], axis="columns")

        # Some metrics can be both inverted and magnitudes (eg rms mag values)
        both = np.logical_and(metric_sets["invert"], metric_sets["mag"])
        # both = 1 + (1-(value-norm) - (1-(norm-norm))) / (1-(norm-norm)) == norm - value
        norm_summary.loc[:, both] = 1.0 - summary.loc[:, both].subtract(
            summary.loc[baseline_comparison, both], axis="columns"
        )

        # Turn the values above into the fractional difference compared with the baseline
        norm_summary.loc[:, :] = 1 + (
            norm_summary.loc[:, :].sub(norm_summary.loc[baseline_comparison, :], axis="columns")
        ).div(norm_summary.loc[baseline_comparison, :], axis="columns")

    # Set the index name
    norm_summary.columns.name = "metric"
    norm_summary.index.name = "run"

    # Make sure we return the rows and metrics in the original order
    norm_summary = norm_summary.loc[runs, used_metrics]

    return norm_summary


def plot_run_metric(
    summary,
    baseline_run=None,
    vertical_quantity="run",
    horizontal_quantity="value",
    run_label_map=None,
    metric_label_map=None,
    metric_set=None,
    ax=None,
    cmap=colorcet.glasbey_hv,
    linestyles=None,
    markers=["o"],
    shade_fraction=0.05,
):
    """Plot normalized metric values as colored points on a cartesian plane.

    Parameters
    ----------
    summary : `pandas.DataFrame`
        Values to be plotted. Should only include runs and metrics that
        should actually appear on the plot.
    baseline_run : `str`
        Name of the run to use as the baseline for normalization (see
        (archive.normalize_metric_summaries).
    vertical_quantity : {'run', 'metric', 'value'}
        Should the run, metric name, or metric value be mapped onto the y axis?
    horizontal_quantity : {'run', 'metric', 'value'}
        Should the run, metric name, or metric value be mapped onto the x axis?
    vwidth : `float`
        The width of the plot, in normalized metrics summary units. (The limits
        of the x axis will be 1-vwidth/2 and 1+width/2).
    run_label_map : mapping
        A python `mapping` between canonical run names and run labels as they
        should appear on plot labels. Use of this option is discouraged, because
        it makes it harder to match plots to data.
        run_label_map could be created by archive.get_runs().loc[these_runs]['brief']
    metric_label_map : mapping
        A python `mapping` between canonical metric names and metric labels
        as they should appear on plot labels. Use this option carefully, because
        it makes it harder to match plots to metric calculation code..
        metric_label_map could be equivalent to metric_set['short_name']
    metric_set : `pandas.DataFrame`
        Metric metadata as returned by `archive.get_metric_sets`
    ax : `matplotlib.axes.Axes`
        The axes on which to plot the data.
    cmap : `matplotlib.colors.ColorMap`
        The color map to use for point colors.
    linestyles : `list`
        A list of matplotlib linestyles to use to connect the lines
    markers : `list`, opt
        A list of matplotlib markers to use to represent the points
    shade_fraction : `float`, opt
        Add a red/green shading to the plot, starting at 1 +/- shade_fraction.
        Set to 0 or None for no shading.

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        The plot figure.
    ax : `matplotilb.axes.Axes`
        The plot axes.


    The run order and metric order (imposed into the summary dataframe passed here as `summary`)
    are important and preserved in the plot. These should be set in the (subset) `summary` dataframe
    passed here; the metric_set is available, but used for normalization and plot styling.
    """

    # If the metric sets we are passed has a multilevel index,
    # get rid of the levels we do not need.
    if metric_set is not None and metric_set.index.nlevels > 1:
        extra_levels = list(set(metric_set.index.names) - set(["metric"]))
        metric_set = metric_set.droplevel(extra_levels).groupby(level="metric").first()

    # Mark whether we have a default, or whether
    # one was specified
    has_linestyle_arg = linestyles is not None
    if not has_linestyle_arg:
        linestyles = [""]

    quantities = set(["run", "metric", "value"])
    assert horizontal_quantity in quantities
    assert vertical_quantity in quantities
    assigned_quantities = set([horizontal_quantity, vertical_quantity])
    color_quantity = (quantities - assigned_quantities).pop()

    # Normalize the summary values, if a baseline was specified
    norm_summary = (
        (
            summary.rename_axis(index="run", columns="metric").copy()
            if baseline_run is None
            else normalize_metric_summaries(baseline_run, summary, metric_sets=metric_set)
        )
        .stack(dropna=False)
        .rename("value")
        .reset_index()
        .rename(columns={"OpsimRun": "run"})
    )
    # Pull original order for metric & runs from summary
    run_order = summary.index.values
    metric_order = summary.columns.values
    if run_label_map is not None:
        run_order = [run_label_map[r] for r in run_order]
        norm_summary["run"] = norm_summary["run"].map(run_label_map)
    if metric_label_map is not None:
        metric_order = [metric_label_map[m] for m in metric_order]
        norm_summary["metric"] = norm_summary["metric"].map(metric_label_map)
        # Create this_metric_set - equivalent to metric_set but with updated names.
        if metric_set is not None:
            this_metric_set = (
                metric_set.drop(columns=["metric"])
                .assign(metric=metric_set["metric"].map(metric_label_map))
                .set_index("metric", drop=False)
            )
        else:
            this_metric_set = None
    else:
        this_metric_set = metric_set

    plot_df = pd.DataFrame(
        {
            "x": norm_summary[horizontal_quantity],
            "y": norm_summary[vertical_quantity],
        }
    )

    if color_quantity == "value":
        warnings.warn("plot_run_metric_mesh is probably a better choice")

        # turn the continuous metric value into a categorical one
        if baseline_run is None:
            # Not normalized, so we have no idea how to scale values.
            # Just let pandas create 7 bins.
            bins = 8
        else:
            bins = (-np.inf, 0, 0.8, 0.95, 1.05, 1.2, 2, np.inf)

        plot_df["color"] = pd.cut(norm_summary[color_quantity], bins)
    else:
        # At this point, color has to be metric or run
        plot_df["color"] = norm_summary[color_quantity]

    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 6))
    else:
        fig = plt.Figure(figsize=(6, 10)) if ax is None else ax.get_figure()

    # make the linestyles and symbols list the same length as cmap
    try:
        num_colors = len(cmap)
        colors = cmap
    except TypeError:
        num_colors = len(cmap.colors)
        colors = cmap.colors

    ls_grow = int(np.ceil(num_colors / len(linestyles)))
    linestyles = (list(linestyles) * ls_grow)[:num_colors]
    marker_grow = int(np.ceil(num_colors / len(markers)))
    markers = (list(markers) * marker_grow)[:num_colors]
    ax.set_prop_cycle(
        cycler.cycler(color=colors) + cycler.cycler(linestyle=linestyles) + cycler.cycler(marker=markers)
    )

    plot_df.set_index("color", inplace=True)

    # 'color' or 'plot_df.index' is now either 'run' or 'metric'
    if color_quantity == "run":
        idx_order = run_order
    if color_quantity == "metric":
        idx_order = metric_order
    for idx in idx_order:  # plot_df.index.unique():
        # Due to weirdness with matplotlib arg handling,
        # make sure we get to pass the style argument
        # as a positional argument, whether or not it is
        # specified.
        # Let the user know why some of their plot values might be disappearing
        # (tends to happen if baseline value is 0 or Nan and plot_df being normalized)
        if vertical_quantity == "value" and np.isinf(plot_df.loc[idx, "y"]).any():
            warnings.warn(f"There are infinite values in the plot of {idx}.")
        if horizontal_quantity == "value" and np.isinf(plot_df.loc[idx, "x"]).any():
            warnings.warn(f"There are infinite values in the plot of {idx}.")
        plot_args = [plot_df.loc[idx, "x"], plot_df.loc[idx, "y"]]
        idx_label = f"{str(idx).strip()}"
        if this_metric_set is not None and idx in this_metric_set.index:
            # Set the style from the metric_set if available
            if "style" in this_metric_set.columns:
                metric_style = this_metric_set.loc[idx, "style"]
                if metric_style is not None:
                    plot_args.append(metric_style)
            # Update the plot label if we inverted the column during normalization
            if "invert" in this_metric_set.columns and baseline_run is not None:
                inv = this_metric_set.loc[idx, "invert"]
                if inv:
                    idx_label = f"1 / {idx_label}"
        ax.plot(*plot_args, label=idx_label)

    if vertical_quantity == "value":
        # Set xlim to be exact length of number of runs
        xlim_new = [0, len(summary) - 1]
        ax.set_xlim(xlim_new)

    if vertical_quantity == "run":
        ylim_new = [0, len(summary) - 1]
        ax.set_ylim(ylim_new)

    if shade_fraction is not None and shade_fraction > 0:
        if vertical_quantity == "value":
            xlim = ax.get_xlim()
            high_shade_bottom = 1 + shade_fraction
            high_shade_top = ax.get_ylim()[1]
            if high_shade_top > high_shade_bottom:
                ax.fill_between(
                    xlim,
                    high_shade_bottom,
                    high_shade_top,
                    color="g",
                    alpha=0.1,
                )

            low_shade_top = 1 - shade_fraction
            low_shade_bottom = ax.get_ylim()[0]
            if low_shade_top > low_shade_bottom:
                ax.fill_between(xlim, low_shade_bottom, low_shade_top, color="r", alpha=0.1)

        elif horizontal_quantity == "value":
            ylim = ax.get_ylim()
            high_shade_left = 1 + shade_fraction
            high_shade_right = ax.get_xlim()[1]
            if high_shade_right > high_shade_left:
                ax.fill_betweenx(
                    ylim,
                    high_shade_left,
                    high_shade_right,
                    color="g",
                    alpha=0.1,
                )

            low_shade_right = 1 - shade_fraction
            low_shade_left = ax.get_xlim()[0]
            if low_shade_right > low_shade_left:
                ax.fill_betweenx(ylim, low_shade_left, low_shade_right, color="r", alpha=0.1)

    if horizontal_quantity in ["run", "metric"]:
        ax.tick_params("x", labelrotation=90)

    ax.legend(bbox_to_anchor=(1.0, 1.0))
    ax.grid()

    return fig, ax


def plot_run_metric_mesh(
    summary,
    metric_set=None,
    baseline_run=None,
    color_range=1,
    run_label_map=None,
    metric_label_map=None,
    ax=None,
    cmap=colorcet.cm["CET_D1A_r"],
):
    """Plot normalized metric values as colored points on a cartesian plane.

    Parameters
    ----------
    summary : `pandas.DataFrame`
        Values to be plotted. Should only include runs and metrics that
        should actually appear on the plot.
    baseline_run : `str`
        Name of the run to use as the baseline for normalization (see
        (archive.normalize_metric_summaries).
    color_range : `float`
        The color range of the plot, in normalized metrics summary
        units. (The color range will be 1-color_range/2 and
        1+color_range/2).
    run_label_map : mapping
        A python `mapping` between canonical run names and run labels as they
        should appear on plot labels.
    metric_label_map : mapping
        A python `mapping` between canonical metric names and metric labels
        as they should appear on plot labels.
    ax : `matplotlib.axes.Axes`
        The axes on which to plot the data.
    cmap : `matplotlib.colors.ColorMap`
        The color map to use for point colors.
        A red/blue diverging color map - CET_D1A_r or CET_D1_r
        A rainbow diverging color map - CET_R3_r

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        The plot figure..
    ax : `matplotilb.axes.Axes`
        The plot axes.

    """
    cols_are_runs = summary.columns.name in RUN_COL_NAMES
    rows_are_metrics = summary.index.name in METRIC_COL_NAMES
    if cols_are_runs or rows_are_metrics:
        summary = summary.T

    # Normalize the summary values, if a baseline was specified
    if baseline_run is not None:
        norm_summary = normalize_metric_summaries(baseline_run, summary, metric_set)
    else:
        norm_summary = summary.rename_axis(index="run", columns="metric").copy()

    if color_range is not None and (isinstance(color_range, float) or isinstance(color_range, int)):
        vmin = 1 - color_range / 2
        vmax = vmin + color_range
    elif isinstance(color_range, list):
        vmin = color_range[0]
        vmax = color_range[1]
    else:
        vmin = norm_summary.min()
        vmax = norm_summary.max()

    norm_values = norm_summary.T.values

    if ax is None:
        fig, ax = plt.subplots(figsize=(6, 6))
    else:
        fig = plt.Figure(figsize=(6, 10)) if ax is None else ax.get_figure()

    y_nums, x_nums = (
        np.arange(norm_summary.shape[1] + 1),
        np.arange(norm_summary.shape[0] + 1),
    )
    im = ax.pcolormesh(
        x_nums,
        y_nums,
        norm_values,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
    )

    ax.set_yticks(np.arange(0.5, norm_summary.shape[1] + 0.5))
    metrics = norm_summary.columns.tolist()
    if metric_label_map is None:
        metric_labels = metrics
    else:
        try:
            # When there are multiple metric families specified,
            # there might be duplicate elements in the
            # metric_label_map. Remove the duplicates.
            metric_label_map = metric_label_map[~metric_label_map.index.duplicated(keep="first")]
        except AttributeError:
            # if metric_label_map is a dict, it won't have
            # an index attrubute, but there isn't any danger
            # if duplicates either
            pass

        # Figure out which metrics get inverted
        if baseline_run is not None and metric_set is not None:
            inverted_metrics = set(metric_set.query("invert").metric.values)
        else:
            inverted_metrics = set()

        metric_labels = [
            f"1/{metric_label_map[m]}" if m in inverted_metrics else metric_label_map[m] for m in metrics
        ]

    ax.set_yticklabels(metric_labels)

    ax.set_xticks(np.arange(0.5, norm_summary.shape[0] + 0.5))
    runs = norm_summary.index.tolist()
    if run_label_map is None:
        run_labels = runs
    else:
        run_labels = [run_label_map[r] for r in runs]
    ax.set_xticklabels(run_labels, rotation="vertical")

    if baseline_run is None:
        fig.colorbar(im, ax=ax, label=None)
    else:
        fig.colorbar(im, ax=ax, label="Fractional difference")

    return fig, ax


def find_family_lines(families, family_list):
    lines = []
    for f in family_list:
        lines += [len(families.loc[f]["run"])]
    lines = np.array(lines).cumsum()
    return lines


# classes

# internal functions & classes



#
# LSST Data Management System
# Copyright 2008, 2009, 2010, 2011, 2012 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#

__all__ = ("Bandpass", "rubin_bandpasses")

import gzip
import os
import warnings

import numpy as np
import scipy.interpolate as interpolate

from rubin_sim.data import get_data_dir

from .physical_parameters import PhysicalParameters
from .sed import Sed


def rubin_bandpasses():
    """Return the standard expected Rubin filter bandpasses."""
    filterlist = ["u", "g", "r", "i", "z", "y"]
    dd = get_data_dir()

    filter_dict = {}
    for f in filterlist:
        filter_dict[f] = Bandpass()
        filter_dict[f].read_throughput(os.path.join(dd, "throughputs", "baseline", f"total_{f}.dat"))
    return filter_dict


class Bandpass:
    """
    Hold and use telescope throughput curves.

    Parameters
    ----------
    wavelen : `np.ndarray`, (N,)
        Wavelength array in nm.
    sb : `np.ndarray`, (N,)
        Throughput array (fraction, 0-1).
    sampling_warning : `float`
        If wavelength sampling lower than this,
        throw a warning because it might provide accurate magnitudes
        due to how magnitudes are calculated in Sed (nm).
    """

    def __init__(self, wavelen=None, sb=None, sampling_warning=0.2):
        self._phys_params = PhysicalParameters()
        self.sampling_warning = sampling_warning
        self.wavelen = None
        self.sb = None
        self.phi = None
        self.bandpassname = None
        if (wavelen is not None) and (sb is not None):
            self.set_bandpass(wavelen, sb)
        return

    def _check_wavelength_sampling(self):
        """Check that the wavelength sampling is above some threshold"""
        if self.wavelen is not None:
            dif = np.diff(self.wavelen)
            if np.max(dif) > self.sampling_warning:
                warnings.warn(
                    "Wavelength sampling of %.1f nm is > %.1f nm" % (np.max(dif), self.sampling_warning)
                    + ", this may not work well"
                    " with a Sed object. Consider resampling with resample_bandpass method."
                )

    def set_bandpass(self, wavelen, sb):
        """
        Populate bandpass data with wavelen/sb arrays.

        Phi set to None.

        Parameters
        ----------
        wavelen : `np.ndarray`, (N,)
            Wavelength array in nm.
        sb : `np.ndarray`, (N,)
            Throughput array (fraction, 0-1).

        Raises
        ------
        ValueError
            Raised if ``wavelen`` and ``sb`` have different length.
        """
        # Check data type.
        if (isinstance(wavelen, np.ndarray) is False) or (isinstance(sb, np.ndarray) is False):
            raise ValueError("Wavelen and sb arrays must be numpy arrays.")
        # Check data matches in length.
        if len(wavelen) != len(sb):
            raise ValueError("Wavelen and sb arrays must have the same length.")

        self.wavelen = np.copy(wavelen)
        self.phi = None
        self.sb = np.copy(sb)
        self.bandpassname = "FromArrays"
        self._check_wavelength_sampling()

    def imsim_bandpass(self, imsimwavelen=500.0, wavelen_min=300, wavelen_max=1150, wavelen_step=0.1):
        """
        Populate bandpass data with sb=0 everywhere except at imsimwavelen.

        Sets wavelen/sb, with grid min/max/step as Parameters.
        """
        # Set up arrays.
        self.wavelen = np.arange(
            wavelen_min,
            wavelen_max + wavelen_step,
            wavelen_step,
            dtype="float",
        )
        self.phi = None
        # Set sb.
        self.sb = np.zeros(len(self.wavelen), dtype="float")
        self.sb[abs(self.wavelen - imsimwavelen) < wavelen_step / 2.0] = 1.0
        self.bandpassname = "IMSIM"
        self._check_wavelength_sampling()

    def read_throughput(self, filename):
        """
        Populate bandpass data with data (wavelen/sb) read from file.

        Sets wavelen/sb. Does NOT set phi.
        """
        # Set self values to None in case of file read error.
        self.wavelen = None
        self.phi = None
        self.sb = None
        # Check for filename error.
        # If given list of filenames,
        # pass to (and return from) read_throughputList.
        if isinstance(filename, list):
            warnings.warn(
                "Was given list of files, instead of a single file. Using read_throughputList instead"
            )
            self.read_throughput_list(component_list=filename)
        # Filename is single file, now try to open file and read data.
        try:
            if filename.endswith(".gz"):
                f = gzip.open(filename, "rt")
            else:
                f = open(filename, "r")
        except IOError:
            try:
                if filename.endswith(".gz"):
                    f = open(filename[:-3], "r")
                else:
                    f = gzip.open(filename + ".gz", "rt")
            except IOError:
                raise IOError("The throughput file %s does not exist" % (filename))
        # The throughput file should have [wavelength(A), throughput(Sb)]
        wavelen = []
        sb = []
        for line in f:
            if line.startswith("#") or line.startswith("$") or line.startswith("!"):
                continue
            values = line.split()
            if len(values) < 2:
                continue
            if (values[0] == "$") or (values[0] == "#") or (values[0] == "!"):
                continue
            wavelen.append(float(values[0]))
            sb.append(float(values[1]))
        f.close()
        self.bandpassname = filename
        # Set up wavelen/sb.
        self.wavelen = np.array(wavelen, dtype="float")
        self.sb = np.array(sb, dtype="float")
        # Check that wavelength is monotonic increasing and
        # non-repeating in wavelength. (Sort on wavelength).
        if len(self.wavelen) != len(np.unique(self.wavelen)):
            raise ValueError("The wavelength values in file %s are non-unique." % (filename))
        # Sort values.
        p = self.wavelen.argsort()
        self.wavelen = self.wavelen[p]
        self.sb = self.sb[p]
        self._check_wavelength_sampling()

    def read_throughput_list(
        self,
        component_list=(
            "detector.dat",
            "lens1.dat",
            "lens2.dat",
            "lens3.dat",
            "m1.dat",
            "m2.dat",
            "m3.dat",
            "atmos_std.dat",
        ),
        root_dir=".",
        wavelen_min=300,
        wavelen_max=1150,
        wavelen_step=0.1,
    ):
        """
        Populate bandpass data by reading from a series of files
        containing wavelen/Sb data.

        Multiplies throughputs (sb) from each file to give a final
        bandpass throughput.
        Sets wavelen/sb, with grid min/max/step as Parameters.
        Does NOT set phi.
        """
        # ComponentList = names of files in that directory.
        # A typical component list of all files to build final component list,
        # including filter, might be:
        # ['detector.dat', 'lens1.dat', 'lens2.dat', 'lens3.dat',
        #  'm1.dat', 'm2.dat', 'm3.dat', 'atmos_std.dat', 'ideal_g.dat']
        #
        # Set up wavelen/sb on grid.
        self.wavelen = np.arange(
            wavelen_min,
            wavelen_max + wavelen_step / 2.0,
            wavelen_step,
            dtype="float",
        )
        self.phi = None
        self.sb = np.ones(len(self.wavelen), dtype="float")
        # Set up a temporary bandpass object to hold data from each file.
        tempbandpass = Bandpass()
        for component in component_list:
            # Read data from file.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                tempbandpass.read_throughput(os.path.join(root_dir, component))
            tempbandpass.resample_bandpass(
                wavelen_min=wavelen_min,
                wavelen_max=wavelen_max,
                wavelen_step=wavelen_step,
            )
            # Multiply self by new sb values.
            self.sb = self.sb * tempbandpass.sb
        self.bandpassname = "".join(component_list)
        self._check_wavelength_sampling()

    def get_bandpass(self):
        wavelen = np.copy(self.wavelen)
        sb = np.copy(self.sb)
        return wavelen, sb

    # utilities

    def check_use_self(self, wavelen, sb):
        """
        Simple utility to check if should be using self.wavelen/sb
        or passed arrays.

        Useful for other methods in this class.
        Also does data integrity check on wavelen/sb if not self.
        """
        update_self = False
        if (wavelen is None) or (sb is None):
            # Then one of the arrays was not passed - check if true for both.
            if (wavelen is not None) or (sb is not None):
                # Then only one of the arrays was passed - raise exception.
                raise ValueError("Must either pass *both* wavelen/sb pair, or use self defaults")
            # Okay, neither wavelen or sb was passed in - using self only.
            update_self = True
        else:
            # Both of the arrays were passed in - check their validity.
            if (isinstance(wavelen, np.ndarray) is False) or (isinstance(sb, np.ndarray) is False):
                raise ValueError("Must pass wavelen/sb as numpy arrays")
            if len(wavelen) != len(sb):
                raise ValueError("Must pass equal length wavelen/sb arrays")
        return update_self

    def resample_bandpass(
        self,
        wavelen=None,
        sb=None,
        wavelen_min=300,
        wavelen_max=1150,
        wavelen_step=0.1,
    ):
        """
        Resamples wavelen/sb (or self.wavelen/sb) onto grid
        defined by min/max/step.

        Either returns wavelen/sb (if given those arrays) or
        updates wavelen / Sb in self.  If updating self, resets phi to None.
        """
        # Check wavelength limits.
        wavelen_min = wavelen_min
        wavelen_max = wavelen_max
        wavelen_step = wavelen_step
        # Is method acting on self.wavelen/sb or passed in wavelen/sb?
        update_self = (wavelen is None) & (sb is None)
        if update_self:
            wavelen = self.wavelen
            sb = self.sb
        # Now, on with the resampling.
        if (wavelen.min() > wavelen_max) or (wavelen.max() < wavelen_min):
            raise Exception("No overlap between known wavelength range and desired wavelength range.")
        # Set up gridded wavelength.
        wavelen_grid = np.arange(wavelen_min, wavelen_max + wavelen_step / 2.0, wavelen_step, dtype="float")
        # Do the interpolation of wavelen/sb onto the grid.
        # (note wavelen/sb type failures will die here).
        f = interpolate.interp1d(wavelen, sb, fill_value=0, bounds_error=False)
        sb_grid = f(wavelen_grid)
        # Update self values if necessary.
        if update_self:
            self.phi = None
            self.wavelen = wavelen_grid
            self.sb = sb_grid
            return
        self._check_wavelength_sampling()
        return wavelen_grid, sb_grid

    # more complicated bandpass functions

    def sb_tophi(self):
        """
        Calculate and set phi - the normalized system response.

        This function only updates self.phi.
        """
        # The definition of phi = (Sb/wavelength)/\int(Sb/wavelength)dlambda.
        self.phi = self.sb / self.wavelen
        # Normalize phi so that the integral of phi is 1.
        norm = np.trapz(self.phi, x=self.wavelen)
        self.phi = self.phi / norm
        return

    def multiply_throughputs(self, wavelen_other, sb_other):
        """
        Multiply self.sb by another wavelen/sb pair, return wavelen/sb arrays.

        The returned arrays will be gridded like this bandpass.
        This method does not affect self.
        """
        # Resample wavelen_other/sb_other to match this bandpass.
        if not np.all(self.wavelen == wavelen_other):
            wavelen_other, sb_other = self.resample_bandpass(
                wavelen=wavelen_other,
                sb=sb_other,
                wavelen_min=self.wavelen.min(),
                wavelen_max=self.wavelen.max(),
                wavelen_step=self.wavelen[1] - self.wavelen[0],
            )
        # Make new memory copy of wavelen.
        wavelen_new = np.copy(self.wavelen)
        # Calculate new transmission - this is also new memory.
        sb_new = self.sb * sb_other
        return wavelen_new, sb_new

    def calc_zp_t(self, photometric_parameters):
        """
        Calculate the instrumental zeropoint for a bandpass.

        Parameters
        ----------
        photometric_parameters : `PhotometricParameters`
            Details about the photometric response of the telescope.
            Defaults to LSST values.
        """
        # ZP_t is the magnitude of a (F_nu flat) source which produced
        # 1 count per second.
        # This is often also known as the 'instrumental zeropoint'.
        # Set gain to 1 if want to explore photo-electrons rather than adu.
        # The typical LSST exposure time is 15s and this is default here,
        # but typical zp_t definition is for 1s.
        # SED class uses flambda in ergs/cm^2/s/nm, so need effarea in cm^2.
        #
        # Check dlambda value for integral.
        self.wavelen[1] - self.wavelen[0]
        # Set up flat source of arbitrary brightness,
        # but where the units of fnu are Jansky (for AB mag zeropoint = -8.9).
        flatsource = Sed()
        flatsource.set_flat_sed()
        adu = flatsource.calc_adu(self, phot_params=photometric_parameters)
        # Scale fnu so that adu is 1 count/expTime.
        flatsource.fnu = flatsource.fnu * (1 / adu)
        # Now need to calculate AB magnitude of the source with this fnu.
        if self.phi is None:
            self.sb_tophi()
        zp_t = flatsource.calc_mag(self)
        return zp_t

    def calc_eff_wavelen(self):
        """
        Calculate effective wavelengths for filters.
        """
        # This is useful for summary numbers for filters.
        # Calculate effective wavelength of filters.
        if self.phi is None:
            self.sb_tophi()
        effwavelenphi = (self.wavelen * self.phi).sum() / self.phi.sum()
        effwavelensb = (self.wavelen * self.sb).sum() / self.sb.sum()
        return effwavelenphi, effwavelensb

    def write_throughput(self, filename, print_header=None, write_phi=False):
        """
        Write throughput to a file.
        """
        # Useful if you build a throughput up from components
        # and need to record the combined value.
        f = open(filename, "w")
        # Print header.
        if print_header is not None:
            if not print_header.startswith("#"):
                print_header = "#" + print_header
            f.write(print_header)
        if write_phi:
            if self.phi is None:
                self.sb_tophi()
            print("# Wavelength(nm)  Throughput(0-1)   Phi", file=f)
        else:
            print("# Wavelength(nm)  Throughput(0-1)", file=f)
        # Loop through data, printing out to file.
        for i in range(0, len(self.wavelen), 1):
            if write_phi:
                print(self.wavelen[i], self.sb[i], self.phi[i], file=f)
            else:
                print(self.wavelen[i], self.sb[i], file=f)
        f.close()



import os
import sqlite3
import unittest

import numpy as np
from rubin_scheduler.data import get_data_dir

import rubin_sim.maf.utils.opsim_utils as opsimUtils


class TestOpsimUtils(unittest.TestCase):
    def test_scale_benchmarks(self):
        """Test scaling the design and stretch benchmarks for the
        length of the run.
        """
        # First test that method returns expected dictionaries.
        for i in ("design", "stretch"):
            benchmark = opsimUtils.scale_benchmarks(10.0, i)
            self.assertIsInstance(benchmark, dict)
            expectedkeys = (
                "Area",
                "nvisitsTotal",
                "nvisits",
                "seeing",
                "skybrightness",
                "singleVisitDepth",
            )
            expectedfilters = ("u", "g", "r", "i", "z", "y")
            for k in expectedkeys:
                self.assertIn(k, benchmark)
            expecteddictkeys = (
                "nvisits",
                "seeing",
                "skybrightness",
                "singleVisitDepth",
            )
            for k in expecteddictkeys:
                for f in expectedfilters:
                    self.assertIn(f, benchmark[k])

    def test_calc_coadded_depth(self):
        """Test the expected coadded depth calculation."""
        benchmark = opsimUtils.scale_benchmarks(10, "design")
        coadd = opsimUtils.calc_coadded_depth(benchmark["nvisits"], benchmark["singleVisitDepth"])
        for f in coadd:
            self.assertLess(coadd[f], 1000)
        singlevisits = {"u": 1, "g": 1, "r": 1, "i": 1, "z": 1, "y": 1}
        coadd = opsimUtils.calc_coadded_depth(singlevisits, benchmark["singleVisitDepth"])
        for f in coadd:
            self.assertAlmostEqual(coadd[f], benchmark["singleVisitDepth"][f])

    def test_get_sim_data(self):
        """Test that we can get simulation data"""
        database_file = os.path.join(get_data_dir(), "tests", "example_dbv1.7_0yrs.db")
        dbcols = ["fieldRA", "fieldDec", "note"]
        sql = "night < 10"
        full_sql = "SELECT fieldRA, fieldDec, note FROM observations where night < 10;"
        # Check that we get data the usual way
        data = opsimUtils.get_sim_data(database_file, sql, dbcols)
        assert np.size(data) > 0

        # Check that we can pass a connection object
        con = sqlite3.connect(database_file)
        data = opsimUtils.get_sim_data(con, sql, dbcols)
        con.close()
        assert np.size(data) > 0

        # Check that kwarg overrides sqlconstraint and dbcols
        data = opsimUtils.get_sim_data(database_file, "blah blah", ["nocol"], full_sql_query=full_sql)
        assert np.size(data) > 0

        # Check that bad file raises an error
        with self.assertRaises(FileNotFoundError):
            opsimUtils.get_sim_data("not_a_file.db", sql, ["nocol"])


if __name__ == "__main__":
    unittest.main()



__all__ = ("name_sanitize", "print_dict", "print_simple_dict")

import sys

import numpy as np


def name_sanitize(in_string):
    """
    Convert a string to a more file name (and web) friendly format.

    Parameters
    ----------
    in_string : str
        The input string to be sanitized. Typically these are combinations of metric names and metadata.

    Returns
    -------
    str
        The string after removal/replacement of non-filename friendly characters.
    """
    # Replace <, > and = signs.
    out_string = in_string.replace(">", "gt").replace("<", "lt").replace("=", "eq")
    # Remove single-spaces, strip '.'s and ','s
    out_string = out_string.replace(" ", "_").replace(".", "_").replace(",", "")
    # and remove / and \
    out_string = out_string.replace("/", "_").replace("\\", "_")
    # and remove parentheses
    out_string = out_string.replace("(", "").replace(")", "")
    # Remove ':' and ';"
    out_string = out_string.replace(":", "_").replace(";", "_")
    # Replace '%' and #
    out_string = out_string.replace("%", "_").replace("#", "_")
    # Remove '__'
    while "__" in out_string:
        out_string = out_string.replace("__", "_")
    return out_string


def _myformat(args, delimiter=" "):
    # Generic line formatter to let you specify delimiter between text fields.
    writestring = ""
    # Wrap in a list if something like an int gets passed in
    if not hasattr(args, "__iter__"):
        args = [args]
    for a in args:
        if isinstance(a, list):
            if len(a) > 1:
                ap = ",".join(map(str, a))
            else:
                ap = "".join(map(str, a))
            writestring += "%s%s" % (ap, delimiter)
        else:
            writestring += "%s%s" % (a, delimiter)
    return writestring


def _myformatdict(adict, delimiter=" "):
    # Generic line formatter used for dictionaries.
    writestring = ""
    for k, v in adict.items():
        if isinstance(v, list):
            if len(v) > 1:
                vp = ",".join(map(str, v))
            else:
                vp = "".join(map(str, v))
            writestring += "%s:%s%s" % (k, vp, delimiter)
        else:
            writestring += "%s:%s%s" % (k, v, delimiter)
    return writestring


def print_dict(content, label, filehandle=None, delimiter=" ", _level=0):
    """
    Print dictionaries (and/or nested dictionaries) nicely.
    Can also print other simpler items (such as numpy ndarray) nicely too.
    This is used to print the config files.

    Parameters
    ----------
    content : dict
        The content to pretty print.
    label : str
        A header for this level of the dictionary.
    filename : file
        Output destination. If None, prints to stdout.
    delimiter : str
        User specified delimiter between fields.
    _level : int
        Internal use (controls level of indent).
    """
    # Get set up with basic file output information.
    if filehandle is None:
        filehandle = sys.stdout
    # And set character to use to indent sets of parameters related to a single dictionary.
    baseindent = "%s" % (delimiter)
    indent = ""
    for i in range(_level - 1):
        indent += "%s" % (baseindent)
    # Print data (this is also the termination of the recursion if given nested dictionaries).
    if not isinstance(content, dict):
        if isinstance(content, str) or isinstance(content, float) or isinstance(content, int):
            print("%s%s%s%s" % (indent, label, delimiter, str(content)), file=filehandle)
        else:
            if isinstance(content, np.ndarray):
                if content.dtype.names is not None:
                    print("%s%s%s" % (indent, delimiter, label), file=filehandle)
                    for element in content:
                        print(
                            "%s%s%s%s%s"
                            % (
                                indent,
                                delimiter,
                                indent,
                                delimiter,
                                _myformat(element),
                            ),
                            file=filehandle,
                        )
                else:
                    print(
                        "%s%s%s%s" % (indent, label, delimiter, _myformat(content)),
                        file=filehandle,
                    )
            else:
                print(
                    "%s%s%s%s" % (indent, label, delimiter, _myformat(content)),
                    file=filehandle,
                )
        return
    # Allow user to specify print order of (some or all) items in order via 'keyorder'.
    #  'keyorder' is list stored in the dictionary.
    if "keyorder" in content:
        orderkeys = content["keyorder"]
        # Check keys in 'keyorder' are actually present in dictionary : remove those which aren't.
        missingkeys = set(orderkeys).difference(set(content.keys()))
        for m in missingkeys:
            orderkeys.remove(m)
        otherkeys = sorted(list(set(content.keys()).difference(set(orderkeys))))
        keys = orderkeys + otherkeys
        keys.remove("keyorder")
    else:
        keys = sorted(content.keys())
    # Print data from dictionary.
    print("%s%s%s:" % (indent, delimiter, label), file=filehandle)
    _level += 2
    for k in keys:
        print_dict(content[k], k, filehandle, delimiter, _level)
    _level -= 2


def print_simple_dict(topdict, subkeyorder, filehandle=None, delimiter=" "):
    """
    Print a simple one-level nested dictionary nicely across the screen,
    with one line per top-level key and all sub-level keys aligned.

    Parameters
    ----------
    topdict : dict
        The dictionary to pretty print
    subkeyorder : list of strings
        The order to print the values of the dictionary.
    filehandle : file
        File output object, if None then uses stdout.
    delimiter : str
        User specified delimiter between fields.
    """
    # Get set up with basic file output information.
    if filehandle is None:
        filehandle = sys.stdout
    # Get all sub-level keys.
    subkeys = []
    for key in topdict:
        subkeys += list(topdict[key].keys())
    subkeys = list(set(subkeys))
    # Align subkeys with 'subkeyorder' and then alphabetize any remaining.
    missingkeys = set(subkeyorder).difference(set(subkeys))
    for m in missingkeys:
        subkeyorder.remove(m)
    otherkeys = sorted(list(set(subkeys).difference(set(subkeyorder))))
    subkeys = subkeyorder + otherkeys
    # Print header.
    writestring = "#"
    for s in subkeys:
        writestring += "%s%s" % (s, delimiter)
    print(writestring, file=filehandle)
    # Now go through and print.
    for k in topdict:
        writestring = ""
        for s in subkeys:
            if s in topdict[k]:
                if (
                    isinstance(topdict[k][s], str)
                    or isinstance(topdict[k][s], float)
                    or isinstance(topdict[k][s], int)
                ):
                    writestring += "%s%s" % (topdict[k][s], delimiter)
                elif isinstance(topdict[k][s], dict):
                    writestring += "%s%s" % (
                        _myformatdict(topdict[k][s], delimiter=delimiter),
                        delimiter,
                    )
                else:
                    writestring += "%s%s" % (_myformat(topdict[k][s]), delimiter)
            else:
                writestring += "%s" % (delimiter)
        print(writestring, file=filehandle)



import os
import subprocess

import healpy as hp
import numpy as np
from astropy.io import fits
from lsst.sims.photUtils import Bandpass, Sed
from lsst.sims.utils import angular_separation

## Tools for calling and reading things from the ESO sky model.
# Downloaded and installed from XXX.
# Installing is a major headache, let's hope we don't have to do this again anytime soon

# Run this in the sm-01_mod2 direcory to regenerate the sims_skybrightness save files.


hPlank = 6.626068e-27  # erg s
cLight = 2.99792458e10  # cm/s


def read_eso_output(filename="output/radspec.fits"):
    """Read in the output generated by sm-01_mod2/bin/calcskymodel"""

    fitsfile = fits.open(filename)
    wave = fitsfile[1].data["lam"].copy() * 1e3  # Wavelength to nm
    header = fitsfile[0].header["comment"]
    spec = fitsfile[1].data["flux"].copy()
    # Convert spectra from ph/s/m2/micron/arcsec2 to erg/s/cm2/nm/arcsec2
    spec = spec / (100.0**2) * hPlank * cLight / (wave * 1e-7) / 1e3

    return spec, wave, header


def write_config(
    outfile="config/skymodel_etc.par",
    sm_h=2.64,
    sm_hmin=2.0,
    alt=90.0,
    alpha=0.0,
    rho=180.0,
    altmoon=-90.0,
    moondist=1.0,
    pres=744.0,
    ssa=0.97,
    calcds="N",
    o3column=1.0,
    moonscal=1.0,
    lon_ecl=135.0,
    lat_ecl=90.0,
    emis_str=0.2,
    temp_str=290.0,
    msolflux=130.0,
    season=0,
    time=0,
    vac_air="vac",
    pwv=2.5,
    rtcode="L",
    resol=60000,
    filepath="data",
    inc_moon="N",
    inc_star="N",
    inc_zodi="N",
    inc_therm="N",
    inc_molec="N",
    inc_upper="N",
    inc_glow="N",
):
    f = open(outfile, "w")

    print("sm_h = %f" % sm_h, file=f)
    print("sm_hmin = %f" % sm_hmin, file=f)
    print("alt = %f" % alt, file=f)
    print("alpha = %f" % alpha, file=f)
    print("rho = %f" % rho, file=f)
    print("altmoon = %f" % altmoon, file=f)
    print("moondist = %f" % moondist, file=f)
    print("pres = %f" % pres, file=f)
    print("ssa = %f" % ssa, file=f)
    print("calcds = %s" % calcds, file=f)
    print("o3column = %f" % o3column, file=f)
    print("moonscal = %f" % moonscal, file=f)
    print("lon_ecl = %f" % lon_ecl, file=f)
    print("lat_ecl = %f" % lat_ecl, file=f)
    print("emis_str = %f" % emis_str, file=f)
    print("temp_str = %f" % temp_str, file=f)
    print("msolflux = %f" % msolflux, file=f)
    print("season = %i" % season, file=f)
    print("time = %i" % time, file=f)
    print("vac_air = %s" % vac_air, file=f)
    print("pwv = %f" % pwv, file=f)
    print("rtcode = %s" % rtcode, file=f)
    print("resol = %i" % resol, file=f)
    print("filepath = %s" % filepath, file=f)

    inc = inc_moon + inc_star + inc_zodi + inc_therm + inc_molec + inc_upper + inc_glow

    print("incl = %s" % inc, file=f)

    f.close()


def call_calcskymodel():
    subprocess.run(["bin/calcskymodel"])


def spec2mags(spectra_list, wave):
    # Load LSST filters
    throughPath = os.getenv("LSST_THROUGHPUTS_BASELINE")
    keys = ["u", "g", "r", "i", "z", "y"]

    dtype = [("mags", "float", (6))]
    result = np.zeros(len(spectra_list), dtype=dtype)

    nfilt = len(keys)
    filters = {}
    for filtername in keys:
        bp = np.loadtxt(
            os.path.join(throughPath, "total_" + filtername + ".dat"),
            dtype=list(zip(["wave", "trans"], [float] * 2)),
        )
        tempB = Bandpass()
        tempB.setBandpass(bp["wave"], bp["trans"])
        filters[filtername] = tempB

    filterwave = np.array([filters[f].calcEffWavelen()[0] for f in keys])

    for i, spectrum in enumerate(spectra_list):
        tempSed = Sed()
        tempSed.setSED(wave, flambda=spectrum)
        for j, filtName in enumerate(keys):
            try:
                result["mags"][i][j] = tempSed.calcMag(filters[filtName])
            except:
                pass
    return result, filterwave


def generate_airglow(outDir=None):
    if outDir is None:
        dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
        outDir = os.path.join(dataDir, "ESO_Spectra/Airglow")

    ams = np.array([1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.5, 3.0])
    specs = []

    alts = 90.0 - np.degrees(np.arccos(1.0 / ams))
    airmasses = []
    solar_flux = []

    for alt, am in zip(alts, ams):
        write_config(alt=alt, inc_glow="Y")
        call_calcskymodel()
        spec, wave, header = read_eso_output()
        specs.append(spec)
        airmasses.append(am)
        # Not doing a range of these this time. I suppose I could.
        solar_flux.append(130)

    mags, filterwave = spec2mags(specs, wave)
    nwave = wave.size
    nspec = len(specs)

    dtype = [
        ("airmass", "float"),
        ("solar_flux", "float"),
        ("spectra", "float", (nwave)),
        ("mags", "float", (6)),
    ]
    spectra = np.zeros(nspec, dtype=dtype)
    spectra["airmass"] = airmasses
    spectra["solar_flux"] = solar_flux
    spectra["spectra"] = specs
    spectra["mags"] = mags["mags"]

    spectra.sort(order=["airmass", "solar_flux"])

    np.savez(
        os.path.join(outDir, "airglowSpectra.npz"),
        wave=wave,
        spec=spectra,
        filterWave=filterwave,
    )


def generate_loweratm(outDir=None):
    if outDir is None:
        dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
        outDir = os.path.join(dataDir, "ESO_Spectra/LowerAtm")

    ams = np.array([1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.5, 3.0])
    specs = []

    alts = 90.0 - np.degrees(np.arccos(1.0 / ams))
    airmasses = []
    nightTimes = []

    for alt, am in zip(alts, ams):
        write_config(alt=alt, inc_molec="Y")
        call_calcskymodel()
        spec, wave, header = read_eso_output()
        specs.append(spec)
        airmasses.append(am)
        # Not doing a range of these this time. I suppose I could.
        nightTimes.append(0)

    mags, filterwave = spec2mags(specs, wave)
    nwave = wave.size
    nspec = len(specs)

    dtype = [
        ("airmass", "float"),
        ("nightTimes", "float"),
        ("spectra", "float", (nwave)),
        ("mags", "float", (6)),
    ]
    spectra = np.zeros(nspec, dtype=dtype)
    spectra["airmass"] = airmasses
    spectra["nightTimes"] = nightTimes
    spectra["spectra"] = specs
    spectra["mags"] = mags["mags"]

    spectra.sort(order=["airmass", "nightTimes"])

    np.savez(
        os.path.join(outDir, "Spectra.npz"),
        wave=wave,
        spec=spectra,
        filterWave=filterwave,
    )


def merged_spec():
    dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
    outDir = os.path.join(dataDir, "ESO_Spectra/MergedSpec")

    # A large number of the background components only depend on Airmass, so we can merge those together

    npzs = [
        "LowerAtm/Spectra.npz",
        "ScatteredStarLight/scatteredStarLight.npz",
        "UpperAtm/Spectra.npz",
    ]
    files = [os.path.join(dataDir, "ESO_Spectra", npz) for npz in npzs]
    temp = np.load(files[0])
    wave = temp["wave"].copy()
    spec = temp["spec"].copy()
    spec["spectra"] = spec["spectra"] * 0.0
    spec["mags"] = spec["mags"] * 0.0

    for filename in files:
        restored = np.load(filename)
        spec["spectra"] += restored["spec"]["spectra"]
        flux = 10.0 ** (-0.4 * (restored["spec"]["mags"] - np.log10(3631.0)))
        flux[np.where(restored["spec"]["mags"] == 0.0)] = 0.0
        spec["mags"] += flux

    spec["mags"] = -2.5 * np.log10(spec["mags"]) + np.log10(3631.0)

    np.savez(
        os.path.join(outDir, "merged_spec.npz"),
        spec=spec,
        wave=wave,
        filterWave=temp["filterWave"],
    )


def generate_moon(outDir=None):
    if outDir is None:
        dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
        outDir = os.path.join(dataDir, "ESO_Spectra/Moon")

    nside = 4
    hpids = np.arange(hp.nside2npix(nside))
    lat, az = hp.pix2ang(nside, hpids)
    alt = np.pi / 2.0 - lat
    airmass = 1.0 / np.cos(np.pi / 2.0 - alt)

    # Only need low airmass and then 1/2 to sky
    good = np.where((az >= 0) & (az <= np.pi) & (airmass <= 3.01) & (airmass >= 1.0))
    airmass = airmass[good]
    alt = np.degrees(alt[good])
    az = np.degrees(az[good])
    hpids = hpids[good]

    moonSunSeps = np.array([0.0, 15, 30, 45, 60, 75, 90, 105, 120, 135, 150, 165, 180], dtype=float)
    moonAlts = np.array([-1.0, 0, 15, 30, 45, 60, 75, 90], dtype=float)

    specs = []
    final_moonSunSep = []
    final_hpid = []
    final_moonAlt = []

    for moonSunSep in moonSunSeps:
        for moonAlt in moonAlts:
            angDists = angular_separation(0.0, moonAlt, az, alt)
            for salt, saz, am, angDist, hpid in zip(alt, az, airmass, angDists, hpids):
                write_config(
                    alpha=moonSunSep,
                    alt=salt,
                    altmoon=moonAlt,
                    rho=angDist,
                    inc_moon="Y",
                )
                call_calcskymodel()
                spec, wave, header = read_eso_output()
                specs.append(spec)
                final_moonSunSep.append(moonSunSep)
                final_hpid.append(hpid)
                final_moonAlt.append(moonAlt)

    mags, filterwave = spec2mags(specs, wave)
    nwave = wave.size
    nspec = len(specs)

    dtype = [
        ("hpid", "int"),
        ("moonAltitude", "float"),
        ("moonSunSep", "float"),
        ("spectra", "float", (nwave)),
        ("mags", "float", (6)),
    ]
    spectra = np.zeros(nspec, dtype=dtype)
    spectra["hpid"] = final_hpid
    spectra["moonAltitude"] = final_moonAlt
    spectra["moonSunSep"] = final_moonSunSep
    spectra["spectra"] = specs
    spectra["mags"] = mags["mags"]

    spectra.sort(order=["moonSunSep", "moonAltitude", "hpid"])

    nbreak = 5
    indices = np.linspace(0, spectra.size, nbreak + 1, dtype=int)

    for i in np.arange(nbreak):
        np.savez(
            os.path.join(outDir, "moonSpectra_" + str(i) + ".npz"),
            wave=wave,
            spec=spectra[indices[i] : indices[i + 1]],
            filterWave=filterwave,
        )


def generate_scatteredStar(outDir=None):
    if outDir is None:
        dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
        outDir = os.path.join(dataDir, "ESO_Spectra/ScatteredStarLight")

    ams = np.array([1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.5, 3.0])
    specs = []

    alts = 90.0 - np.degrees(np.arccos(1.0 / ams))
    airmasses = []
    nightTimes = []

    for alt, am in zip(alts, ams):
        write_config(alt=alt, inc_star="Y")
        call_calcskymodel()
        spec, wave, header = read_eso_output()
        specs.append(spec)
        airmasses.append(am)
        # Not doing a range of these this time. I suppose I could.
        nightTimes.append(0)

    mags, filterwave = spec2mags(specs, wave)
    nwave = wave.size
    nspec = len(specs)

    dtype = [
        ("airmass", "float"),
        ("nightTimes", "float"),
        ("spectra", "float", (nwave)),
        ("mags", "float", (6)),
    ]
    spectra = np.zeros(nspec, dtype=dtype)
    spectra["airmass"] = airmasses
    spectra["nightTimes"] = nightTimes
    spectra["spectra"] = specs
    spectra["mags"] = mags["mags"]

    spectra.sort(order=["airmass", "nightTimes"])

    np.savez(
        os.path.join(outDir, "scatteredStarLight.npz"),
        wave=wave,
        spec=spectra,
        filterWave=filterwave,
    )


def generate_upperatm(outDir=None):
    if outDir is None:
        dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
        outDir = os.path.join(dataDir, "ESO_Spectra/UpperAtm")

    ams = np.array([1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.5, 3.0])
    specs = []

    alts = 90.0 - np.degrees(np.arccos(1.0 / ams))
    airmasses = []
    nightTimes = []

    for alt, am in zip(alts, ams):
        write_config(alt=alt, inc_upper="Y")
        call_calcskymodel()
        spec, wave, header = read_eso_output()
        specs.append(spec)
        airmasses.append(am)
        # Not doing a range of these this time. I suppose I could.
        nightTimes.append(0)

    mags, filterwave = spec2mags(specs, wave)
    nwave = wave.size
    nspec = len(specs)

    dtype = [
        ("airmass", "float"),
        ("nightTimes", "float"),
        ("spectra", "float", (nwave)),
        ("mags", "float", (6)),
    ]
    spectra = np.zeros(nspec, dtype=dtype)
    spectra["airmass"] = airmasses
    spectra["nightTimes"] = nightTimes
    spectra["spectra"] = specs
    spectra["mags"] = mags["mags"]

    spectra.sort(order=["airmass", "nightTimes"])

    np.savez(
        os.path.join(outDir, "Spectra.npz"),
        wave=wave,
        spec=spectra,
        filterWave=filterwave,
    )


def generate_zodi(outDir=None):
    if outDir is None:
        dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")
        outDir = os.path.join(dataDir, "ESO_Spectra/Zodiacal")

    ams = np.array([1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.5, 3.0])
    specs = []
    final_hpid = []
    nside = 4
    hpids = np.arange(hp.nside2npix(nside))
    theta, phi = hp.pix2ang(nside, hpids)
    lon = np.degrees(phi)
    lon[np.where(lon > 180.0)] = lon[np.where(lon > 180.0)] - 360.0
    lat = np.degrees(theta) - 90.0

    alts = 90.0 - np.degrees(np.arccos(1.0 / ams))
    airmasses = []

    for la, lo, hpi in zip(lat, lon, hpids):
        for alt, am in zip(alts, ams):
            write_config(alt=alt, lon_ecl=lo, lat_ecl=la, inc_zodi="Y")
            call_calcskymodel()
            spec, wave, header = read_eso_output()
            specs.append(spec)
            airmasses.append(am)
            final_hpid.append(hpi)

    mags, filterwave = spec2mags(specs, wave)
    nwave = wave.size
    nspec = len(specs)

    dtype = [
        ("airmass", "float"),
        ("hpid", "int"),
        ("spectra", "float", (nwave)),
        ("mags", "float", (6)),
    ]
    spectra = np.zeros(nspec, dtype=dtype)
    spectra["airmass"] = airmasses
    spectra["hpid"] = final_hpid
    spectra["spectra"] = specs
    spectra["mags"] = mags["mags"]

    spectra.sort(order=["airmass", "hpid"])

    # span this over multiple files to store in github
    nbreak = 3
    indices = np.linspace(0, spectra.size, nbreak + 1, dtype=int)

    for i in np.arange(nbreak):
        np.savez(
            os.path.join(outDir, "zodiacalSpectra_" + str(i) + ".npz"),
            wave=wave,
            spec=spectra[indices[i] : indices[i + 1]],
            filterWave=filterwave,
        )


if __name__ == "__main__":
    generate_airglow()
    generate_loweratm()
    generate_moon()
    generate_scatteredStar()
    generate_upperatm()
    generate_zodi()
    merged_spec()



__all__ = ("LsqrSolver",)

import numpy as np
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import lsqr

# Modified from /astro/store/scratch/tmp/yoachim/Research/LSST/Parallel_Solver
# and then from https://github.com/lsst-sims/legacy_sims_selfcal


class LsqrSolver:
    """
    Class to solve self-calibration

    Parameters
    ----------
    observaitons : np.array
        A numpy array of the observations. Should have columns id, patch_id, observed_mag, mag_uncert
    patch_out : str ("solved_patches.npz")
        Output file for patch solutions, can be set to None
    star_out : str ("solved_stars.npz")
        Output file for star solutions, can be set to None
    atol : float (1e-8)
        Tolerance passed to lsqr
    btol : float (1e-8)
        Tolerance passed to lsqr
    iter_lim : int (None)
        Iteration limit passed to lsqr
    show : bool (False)
        Should the lsqr solver print some iteration logs (False).
    """

    def __init__(
        self,
        observations,
        atol=1e-8,
        btol=1e-8,
        iter_lim=None,
        show=False,
    ):
        self.atol = atol
        self.btol = btol
        self.iter_lim = iter_lim
        self.observations = observations
        self.show = show

    def run(self):
        """clean data, solve matrix, write solution out."""
        self.clean_data()
        self.solve_matrix()

    def clean_data(self):
        """
        Remove observations that can't contribute to a solution.
        Index remaining stars and patches so they are continuous.
        """
        n_start = 1.0
        n_end = 0.0
        while n_start != n_end:
            n_start = self.observations.size
            self.observations.sort(order="id")
            # Remove observations if the star was only observed once
            good = np.where(
                (self.observations["id"] - np.roll(self.observations["id"], 1))
                * (self.observations["id"] - np.roll(self.observations["id"], -1))
                == 0
            )
            self.observations = self.observations[good]

            # Remove patches with only one star
            self.observations.sort(order="patch_id")

            good = np.where(
                (self.observations["patch_id"] - np.roll(self.observations["patch_id"], 1))
                * (self.observations["patch_id"] - np.roll(self.observations["patch_id"], -1))
                == 0
            )
            self.observations = self.observations[good]
            n_end = self.observations.size

        self.observations.sort(order="patch_id")

        self.patches = np.unique(self.observations["patch_id"])
        n_patches = np.size(self.patches)
        self.n_patches = n_patches
        self.n_patches = n_patches
        patches_index = np.arange(n_patches)
        left = np.searchsorted(self.observations["patch_id"], self.patches)
        right = np.searchsorted(self.observations["patch_id"], self.patches, side="right")
        for i in range(np.size(left)):
            self.observations["patch_id"][left[i] : right[i]] = patches_index[i]

        # Convert id to continuous running index to keep matrix as small as possible
        self.observations.sort(order="id")

        self.stars = np.unique(self.observations["id"])
        n_stars = np.size(self.stars)
        self.n_stars = np.size(self.stars)
        stars_index = np.arange(1, n_stars + 1)
        left = np.searchsorted(self.observations["id"], self.stars)
        right = np.searchsorted(self.observations["id"], self.stars, side="right")
        for i in range(np.size(left)):
            self.observations["id"][left[i] : right[i]] = stars_index[i]

    def solve_matrix(self):
        n_obs = np.size(self.observations)
        # construct sparse matrix
        # A = lil_matrix((nPatches+nStars,np.size(observations['patch_id'])))
        row = np.arange(n_obs)
        row = np.append(row, row)
        col = np.append(
            self.observations["patch_id"],
            self.observations["id"] + np.max(self.observations["patch_id"]),
        )
        # data = np.append(np.ones(nObs),1./observations['mag_uncert'])
        data = 1.0 / self.observations["mag_uncert"]
        data = np.append(data, data)
        b = (
            self.observations["observed_mag"] / self.observations["mag_uncert"]
        )  # maybe do this in place earlier?  then I can just delete parts of observations earlier to save total memory

        # blast away data now that we have the matrix constructed
        del self.observations

        A = coo_matrix((data, (row, col)), shape=(n_obs, self.n_patches + self.n_stars))
        A = A.tocsr()
        # solve Ax = b
        self.solution = lsqr(A, b, show=self.show, atol=self.atol, btol=self.btol, iter_lim=self.iter_lim)

    def return_solution(self):
        """
        Returns
        -------
        np.array with patch zeropoints and star best-fit mags.
        """
        patches = np.empty(self.patches.size, dtype=list(zip(["patch_id", "zp"], [int, float])))
        patches["patch_id"] = self.patches
        patches["zp"] = self.solution[0][0 : self.n_patches]

        stars = np.empty(self.stars.size, dtype=list(zip(["id", "fit_mag"], [int, float])))
        stars["id"] = self.stars  # should match the input ID.
        stars["fit_mag"] = self.solution[0][self.n_patches :]

        return patches, stars



import os

import lsst.sims.photUtils.Sed as Sed
import numpy as np

dataDir = os.getenv("SIMS_SKYBRIGHTNESS_DATA_DIR")

data = np.genfromtxt(
    os.path.join(dataDir, "solarSpec/solarSpec.dat"),
    dtype=list(zip(["microns", "Irr"], [float] * 2)),
)
# data['Irr'] = data['Irr']*1 #convert W/m2/micron to erg/s/cm2/nm (HA, it's the same!)

sun = Sed()
sun.setSED(data["microns"] * 1e3, flambda=data["Irr"])

# Match the wavelenth spacing and range to the ESO spectra
airglowSpec = np.load(os.path.join(dataDir, "ESO_Spectra/Airglow/airglowSpectra.npz"))
sun.resampleSED(wavelen_match=airglowSpec["wave"])

np.savez(os.path.join(dataDir, "solarSpec/solarSpec.npz"), wave=sun.wavelen, spec=sun.flambda)



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics


class TestCadenceMetrics(unittest.TestCase):
    def test_phase_gap_metric(self):
        """
        Test the phase gap metric
        """
        data = np.zeros(10, dtype=list(zip(["observationStartMJD"], [float])))
        data["observationStartMJD"] += np.arange(10) * 0.25

        pgm = metrics.PhaseGapMetric(n_periods=1, period_min=0.5, period_max=0.5)
        metric_val = pgm.run(data)

        mean_gap = pgm.reduce_mean_gap(metric_val)
        median_gap = pgm.reduce_median_gap(metric_val)
        worst_period = pgm.reduce_worst_period(metric_val)
        largest_gap = pgm.reduce_largest_gap(metric_val)

        self.assertEqual(mean_gap, 0.5)
        self.assertEqual(median_gap, 0.5)
        self.assertEqual(worst_period, 0.5)
        self.assertEqual(largest_gap, 0.5)

        pgm = metrics.PhaseGapMetric(n_periods=2, period_min=0.25, period_max=0.5)
        metric_val = pgm.run(data)

        mean_gap = pgm.reduce_mean_gap(metric_val)
        median_gap = pgm.reduce_median_gap(metric_val)
        worst_period = pgm.reduce_worst_period(metric_val)
        largest_gap = pgm.reduce_largest_gap(metric_val)

        self.assertEqual(mean_gap, 0.75)
        self.assertEqual(median_gap, 0.75)
        self.assertEqual(worst_period, 0.25)
        self.assertEqual(largest_gap, 1.0)

    def test_template_exists(self):
        """
        Test the TemplateExistsMetric.
        """
        names = ["finSeeing", "observationStartMJD"]
        types = [float, float]
        data = np.zeros(10, dtype=list(zip(names, types)))
        data["finSeeing"] = [2.0, 2.0, 3.0, 1.0, 1.0, 1.0, 0.5, 1.0, 0.4, 1.0]
        data["observationStartMJD"] = np.arange(10)
        slice_point = {"sid": 0}
        # so here we have 4 images w/o good previous templates
        metric = metrics.TemplateExistsMetric(seeing_col="finSeeing")
        result = metric.run(data, slice_point)
        self.assertEqual(result, 6.0 / 10.0)

    def test_uniformity_metric(self):
        names = ["observationStartMJD"]
        types = [float]
        data = np.zeros(100, dtype=list(zip(names, types)))
        metric = metrics.UniformityMetric()
        result1 = metric.run(data)
        # If all the observations are on the 1st day, should be 1
        self.assertEqual(result1, 1)
        data["observationStartMJD"] = data["observationStartMJD"] + 365.25 * 10
        slice_point = {"sid": 0}
        result2 = metric.run(data, slice_point)
        # All on last day should also be 1
        self.assertEqual(result2, 1)
        # Make a perfectly uniform dist
        data["observationStartMJD"] = np.arange(0.0, 365.25 * 10, 365.25 * 10 / 100)
        result3 = metric.run(data, slice_point)
        # Result should be zero for uniform
        np.testing.assert_almost_equal(result3, 0.0)
        # A single obseravtion should give a result of 1
        data = np.zeros(1, dtype=list(zip(names, types)))
        result4 = metric.run(data, slice_point)
        self.assertEqual(result4, 1)

    def test_general_uniformity_metric(self):
        names = ["observationStartMJD"]
        types = [float]
        data = np.zeros(100, dtype=list(zip(names, types)))
        metric = metrics.GeneralUniformityMetric(col="observationStartMJD", min_value=0, max_value=1)
        result1 = metric.run(data)
        # If all the observations are at the minimum value, should be 1
        self.assertEqual(result1, 1)
        data["observationStartMJD"] = data["observationStartMJD"] + 1
        slice_point = {"sid": 0}
        result2 = metric.run(data, slice_point)
        # All on max value should also be 1
        self.assertEqual(result2, 1)
        # Make a perfectly uniform dist
        step = 1 / 99
        data["observationStartMJD"] = np.arange(0.0, 1 + step / 2, step)
        result3 = metric.run(data, slice_point)
        # Result should be zero for uniform
        np.testing.assert_almost_equal(result3, 0.0)
        # Test setup with max_value > 1
        metric = metrics.GeneralUniformityMetric(col="observationStartMJD", min_value=0, max_value=10)
        # Make a perfectly uniform dist covering limited range
        # (0-1 out of 0-10 range)
        data["observationStartMJD"] = np.arange(0.0, 1 + step / 2, step)
        result5 = metric.run(data, slice_point)
        self.assertEqual(result5, 0.9)
        # Test setup with max_value None
        metric = metrics.GeneralUniformityMetric(col="observationStartMJD", min_value=None, max_value=None)
        # Make a perfectly uniform dist
        data["observationStartMJD"] = np.arange(0.0, 1 + step / 2, step)
        result5 = metric.run(data, slice_point)
        self.assertAlmostEqual(result5, 0)
        # A single observation should give a result of 1
        data = np.zeros(1, dtype=list(zip(names, types)))
        result4 = metric.run(data, slice_point)
        self.assertEqual(result4, 1)

    def test_t_gap_metric(self):
        names = ["observationStartMJD"]
        types = [float]
        data = np.zeros(100, dtype=list(zip(names, types)))
        # All 1-day gaps
        data["observationStartMJD"] = np.arange(100)

        metric = metrics.TgapsMetric(bins=np.arange(1, 100, 1))
        result1 = metric.run(data)
        # By default, should all be in first bin
        self.assertEqual(result1[0], data.size - 1)
        self.assertEqual(np.sum(result1), data.size - 1)
        data["observationStartMJD"] = np.arange(0, 200, 2)
        result2 = metric.run(data)
        self.assertEqual(result2[1], data.size - 1)
        self.assertEqual(np.sum(result2), data.size - 1)

        data = np.zeros(4, dtype=list(zip(names, types)))
        data["observationStartMJD"] = [10, 20, 30, 40]
        metric = metrics.TgapsMetric(all_gaps=True, bins=np.arange(1, 100, 10))
        result3 = metric.run(data)
        self.assertEqual(result3[1], 2)
        ngaps = np.math.factorial(data.size - 1)
        self.assertEqual(np.sum(result3), ngaps)

    def test_t_gaps_percent_metric(self):
        names = ["observationStartMJD"]
        types = [float]
        data = np.zeros(100, dtype=list(zip(names, types)))

        # All 1-day gaps
        data["observationStartMJD"] = np.arange(100)
        # All intervals are one day, so should be 100% within the gap specified
        metric = metrics.TgapsPercentMetric(min_time=0.5, max_time=1.5, all_gaps=False)
        result1 = metric.run(data)
        self.assertEqual(result1, 100)

        # All 2 day gaps
        data["observationStartMJD"] = np.arange(0, 200, 2)
        result2 = metric.run(data)
        self.assertEqual(result2, 0)

        # Run with all_gaps = True
        data = np.zeros(4, dtype=list(zip(names, types)))
        data["observationStartMJD"] = [1, 2, 3, 4]
        metric = metrics.TgapsPercentMetric(min_time=0.5, max_time=1.5, all_gaps=True)
        result3 = metric.run(data)
        # This should be 50% -- 3 gaps of 1 day, 2 gaps of 2 days, and
        # 1 gap of 3 days
        self.assertEqual(result3, 50)

    def test_night_gap_metric(self):
        names = ["night"]
        types = [float]
        data = np.zeros(100, dtype=list(zip(names, types)))
        # All 1-day gaps
        data["night"] = np.arange(100)

        metric = metrics.NightgapsMetric(bins=np.arange(1, 100, 1))
        result1 = metric.run(data)
        # By default, should all be in first bin
        self.assertEqual(result1[0], data.size - 1)
        self.assertEqual(np.sum(result1), data.size - 1)
        data["night"] = np.arange(0, 200, 2)
        result2 = metric.run(data)
        self.assertEqual(result2[1], data.size - 1)
        self.assertEqual(np.sum(result2), data.size - 1)

        data = np.zeros(4, dtype=list(zip(names, types)))
        data["night"] = [10, 20, 30, 40]
        metric = metrics.NightgapsMetric(all_gaps=True, bins=np.arange(1, 100, 10))
        result3 = metric.run(data)
        self.assertEqual(result3[1], 2)
        ngaps = np.math.factorial(data.size - 1)
        self.assertEqual(np.sum(result3), ngaps)

        data = np.zeros(6, dtype=list(zip(names, types)))
        data["night"] = [1, 1, 2, 3, 3, 5]
        metric = metrics.NightgapsMetric(bins=np.arange(0, 5, 1))
        result4 = metric.run(data)
        self.assertEqual(result4[0], 0)
        self.assertEqual(result4[1], 2)
        self.assertEqual(result4[2], 1)

    def test_n_visits_per_night_metric(self):
        names = ["night"]
        types = [float]
        data = np.zeros(100, dtype=list(zip(names, types)))
        # One visit per night.
        data["night"] = np.arange(100)

        bins = np.arange(0, 5, 1)
        metric = metrics.NVisitsPerNightMetric(bins=bins)
        result = metric.run(data)
        # All nights have one visit.
        expected_result = np.zeros(len(bins) - 1, dtype=int)
        expected_result[1] = len(data)
        np.testing.assert_array_equal(result, expected_result)

        data["night"] = np.floor(np.arange(0, 100) / 2)
        result = metric.run(data)
        expected_result = np.zeros(len(bins) - 1, dtype=int)
        expected_result[2] = len(data) / 2
        np.testing.assert_array_equal(result, expected_result)

    def test_rapid_revisit_uniformity_metric(self):
        data = np.zeros(100, dtype=list(zip(["observationStartMJD"], [float])))
        # Uniformly distribute time _differences_ between 0 and 100
        dtimes = np.arange(100)
        data["observationStartMJD"] = dtimes.cumsum()
        # Set up "rapid revisit" metric to look for visits between 5 and 25
        metric = metrics.RapidRevisitUniformityMetric(d_tmin=5, d_tmax=55, min_nvisits=50)
        result = metric.run(data)
        # This should be uniform.
        self.assertLess(result, 0.1)
        self.assertGreaterEqual(result, 0)
        # Set up non-uniform distribution of time differences
        dtimes = np.zeros(100) + 5
        data["observationStartMJD"] = dtimes.cumsum()
        result = metric.run(data)
        self.assertGreaterEqual(result, 0.5)
        dtimes = np.zeros(100) + 15
        data["observationStartMJD"] = dtimes.cumsum()
        result = metric.run(data)
        self.assertGreaterEqual(result, 0.5)
        """
        # Let's see how much dmax/result can vary
        resmin = 1
        resmax = 0
        rng = np.random.RandomState(88123100)
        for i in range(10000):
            dtimes = rng.rand(100)
            data['observationStartMJD'] = dtimes.cumsum()
            metric = metrics.RapidRevisitUniformityMetric(d_tmin=0.1,
                                d_tmax=0.8, min_nvisits=50)
            result = metric.run(data)
            resmin = np.min([resmin, result])
            resmax = np.max([resmax, result])
        print("RapidRevisitUniformity .. range", resmin, resmax)
        """

    def test_rapid_revisit_metric(self):
        data = np.zeros(100, dtype=list(zip(["observationStartMJD"], [float])))
        dtimes = np.arange(100) / 24.0 / 60.0
        data["observationStartMJD"] = dtimes.cumsum()
        # Set metric parameters to the actual N1/N2 values for these dtimes.
        metric = metrics.RapidRevisitMetric(
            d_tmin=40.0 / 60.0 / 60.0 / 24.0,
            d_tpairs=20.0 / 60.0 / 24.0,
            d_tmax=30.0 / 60.0 / 24.0,
            min_n1=19,
            min_n2=29,
        )
        result = metric.run(data)
        self.assertEqual(result, 1)
        # Set metric parameters to > N1/N2 values, to see it return 0.
        metric = metrics.RapidRevisitMetric(
            d_tmin=40.0 / 60.0 / 60.0 / 24.0,
            d_tpairs=20.0 / 60.0 / 24.0,
            d_tmax=30.0 / 60.0 / 24.0,
            min_n1=30,
            min_n2=50,
        )
        result = metric.run(data)
        self.assertEqual(result, 0)
        # Test with single value data.
        data = np.zeros(1, dtype=list(zip(["observationStartMJD"], [float])))
        result = metric.run(data)
        self.assertEqual(result, 0)

    def test_n_revisits_metric(self):
        data = np.zeros(100, dtype=list(zip(["observationStartMJD"], [float])))
        dtimes = np.arange(100) / 24.0 / 60.0
        data["observationStartMJD"] = dtimes.cumsum()
        metric = metrics.NRevisitsMetric(d_t=50.0)
        result = metric.run(data)
        self.assertEqual(result, 50)
        metric = metrics.NRevisitsMetric(d_t=50.0, normed=True)
        result = metric.run(data)
        self.assertEqual(result, 0.5)

    def test_transient_metric(self):
        names = ["observationStartMJD", "fiveSigmaDepth", "filter"]
        types = [float, float, "<U1"]

        ndata = 100
        data_slice = np.zeros(ndata, dtype=list(zip(names, types)))
        data_slice["observationStartMJD"] = np.arange(ndata)
        data_slice["fiveSigmaDepth"] = 25
        data_slice["filter"] = "g"

        metric = metrics.TransientMetric(survey_duration=ndata / 365.25)

        # Should detect everything
        self.assertEqual(metric.run(data_slice), 1.0)

        # Double to survey duration, should now only detect half
        metric = metrics.TransientMetric(survey_duration=ndata / 365.25 * 2)
        self.assertEqual(metric.run(data_slice), 0.5)

        # Set half of the m5 of the observations very bright,
        # so kill another half.
        data_slice["fiveSigmaDepth"][0:50] = 20
        self.assertEqual(metric.run(data_slice), 0.25)

        data_slice["fiveSigmaDepth"] = 25
        # Demand lots of early observations
        metric = metrics.TransientMetric(peak_time=0.5, n_pre_peak=3, survey_duration=ndata / 365.25)
        self.assertEqual(metric.run(data_slice), 0.0)

        # Demand a reasonable number of early observations
        metric = metrics.TransientMetric(peak_time=2, n_pre_peak=2, survey_duration=ndata / 365.25)
        self.assertEqual(metric.run(data_slice), 1.0)

        # Demand multiple filters
        metric = metrics.TransientMetric(n_filters=2, survey_duration=ndata / 365.25)
        self.assertEqual(metric.run(data_slice), 0.0)

        data_slice["filter"] = ["r", "g"] * 50
        self.assertEqual(metric.run(data_slice), 1.0)

        # Demad too many observation per light curve
        metric = metrics.TransientMetric(n_per_lc=20, survey_duration=ndata / 365.25)
        self.assertEqual(metric.run(data_slice), 0.0)

        # Test both filter and number of LC samples
        metric = metrics.TransientMetric(n_filters=2, n_per_lc=3, survey_duration=ndata / 365.25)
        self.assertEqual(metric.run(data_slice), 1.0)

    def test_season_length_metric(self):
        times = np.arange(0, 3650, 10)
        data = np.zeros(
            len(times),
            dtype=list(zip(["observationStartMJD", "visitExposureTime"], [float, float])),
        )
        data["observationStartMJD"] = times
        data["visitExposureTime"] = 30.0
        metric = metrics.SeasonLengthMetric(reduce_func=np.median)
        slice_point = {"ra": 0}
        result = metric.run(data, slice_point)
        self.assertEqual(result, 350)
        times = np.arange(0, 3650, 365)
        data = np.zeros(
            len(times) * 2,
            dtype=list(zip(["observationStartMJD", "visitExposureTime"], [float, float])),
        )
        data["observationStartMJD"][0 : len(times)] = times
        data["observationStartMJD"][len(times) :] = times + 10
        data["observationStartMJD"] = np.sort(data["observationStartMJD"])
        data["visitExposureTime"] = 30.0
        metric = metrics.SeasonLengthMetric(reduce_func=np.median)
        slice_point = {"ra": 0}
        result = metric.run(data, slice_point)
        self.assertEqual(result, 10)
        times = np.arange(0, 3650 - 365, 365)
        data = np.zeros(
            len(times) * 2,
            dtype=list(zip(["observationStartMJD", "visitExposureTime"], [float, float])),
        )
        data["observationStartMJD"][0 : len(times)] = times
        data["observationStartMJD"][len(times) :] = times + 10
        data["observationStartMJD"] = np.sort(data["observationStartMJD"])
        data["visitExposureTime"] = 30.0
        metric = metrics.SeasonLengthMetric(reduce_func=np.size)
        slice_point = {"ra": 0}
        result = metric.run(data, slice_point)
        self.assertEqual(result, 9)


if __name__ == "__main__":
    unittest.main()



__all__ = ("FilterPairTGapsMetric",)

import numpy as np

from ..metrics.base_metric import BaseMetric


class FilterPairTGapsMetric(BaseMetric):
    """
    figure of merit to measure the coverage the time gaps in same and different filter pairs;
    FoM is defined as sum of Nv / standard deviation after a clip;
    Parameters:
        colname: list, ['observationStartMJD', 'filter', 'fiveSigmaDepth']
        fltpairs: filter pair, default ['uu', 'ug', 'ur', 'ui', 'uz','uy',
                                        'gg', 'gr', 'gi', 'gz', 'gy',
                                        'rr', 'ri', 'rz', 'ry',
                                        'ii', 'iz', 'iy',
                                        'zz', 'zy',
                                        'yy']
        mag_lim: list, fiveSigmaDepth threshold each filter, default {'u':18, 'g':18, 'r':18, 'i':18, 'z':18, 'y':18}
        bins_same: np.array, bins to get histogram for same-filter pair ;
        bins_diff: np.array, bins to get histogram for diff-filter pair ;
        nv_clip: number of visits of pairs to clip, std is calculated below nv_clip
        allgaps: boolean, all possible pairs if True, else consider only nearest

    Returns:
        result: sum of fom for all filterpairs,

    """

    def __init__(
        self,
        mjd_col="observationStartMJD",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        fltpairs=[
            "uu",
            "ug",
            "ur",
            "ui",
            "uz",
            "uy",
            "gg",
            "gr",
            "gi",
            "gz",
            "gy",
            "rr",
            "ri",
            "rz",
            "ry",
            "ii",
            "iz",
            "iy",
            "zz",
            "zy",
            "yy",
        ],
        mag_lim={"u": 18, "g": 18, "r": 18, "i": 18, "z": 18, "y": 18},
        bins_same=np.logspace(np.log10(5 / 60 / 60 / 24), np.log10(3650), 50),
        bins_diff=np.logspace(np.log10(5 / 60 / 60 / 24), np.log10(2), 50),
        nv_clip={
            "uu": 30,
            "ug": 30,
            "ur": 30,
            "ui": 30,
            "uz": 30,
            "uy": 30,
            "gg": 69,
            "gr": 30,
            "gi": 30,
            "gz": 30,
            "gy": 30,
            "rr": 344,
            "ri": 30,
            "rz": 30,
            "ry": 30,
            "ii": 355,
            "iz": 30,
            "iy": 30,
            "zz": 282,
            "zy": 30,
            "yy": 288,
        },
        allgaps=True,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.filter_col = filter_col
        self.m5_col = m5_col
        self.fltpairs = fltpairs
        self.mag_lim = mag_lim
        self.bins_same = bins_same
        self.bins_diff = bins_diff

        self.allgaps = allgaps

        # number of visits to clip, default got from 1/10th of baseline_v2.0 WFD
        self.nv_clip = nv_clip

        super().__init__(col=[self.mjd_col, self.filter_col, self.m5_col], **kwargs)

    def _get_d_t(self, data_slice, f0, f1):
        # select
        idx0 = (data_slice[self.filter_col] == f0) & (data_slice[self.m5_col] > self.mag_lim[f0])
        idx1 = (data_slice[self.filter_col] == f1) & (data_slice[self.m5_col] > self.mag_lim[f1])

        time_col0 = data_slice[self.mjd_col][idx0]
        time_col1 = data_slice[self.mjd_col][idx1]

        # time_col0 = time_col0.reshape((len(time_col0), 1))
        # time_col1 = time_col1.reshape((len(time_col1), 1))

        # calculate time gaps matrix
        # diffmat = np.subtract(time_col0, time_col1.T)

        if self.allgaps:
            # collect all time gaps
            if f0 == f1:
                time_col0 = time_col0.reshape((len(time_col0), 1))
                time_col1 = time_col1.reshape((len(time_col1), 1))

                diffmat = np.subtract(time_col0, time_col1.T)
                diffmat = np.abs(diffmat)
                # get only triangle part
                dt_tri = np.tril(diffmat, -1)
                d_t = dt_tri[dt_tri != 0]  # flatten lower triangle
            else:
                # d_t = diffmat.flatten()
                dtmax = np.max(self.bins_diff)  # time gaps window for measure color
                d_t = []
                for time_col in time_col0:
                    time_col_in_window = time_col1[
                        (time_col1 >= (time_col - dtmax)) & (time_col1 <= (time_col + dtmax))
                    ]

                    d_t.append(np.abs(time_col_in_window - time_col))

                d_t = np.concatenate(d_t) if len(d_t) > 0 else np.array(d_t)
        else:
            # collect only nearest
            if f0 == f1:
                # get diagonal ones nearest nonzero, offset=1
                # d_t = np.diagonal(diffmat, offset=1)
                d_t = np.diff(time_col0)
            else:
                time_col0 = data_slice[self.mjd_col][idx0]
                time_col1 = data_slice[self.mjd_col][idx1]

                time_col0 = time_col0.reshape((len(time_col0), 1))
                time_col1 = time_col1.reshape((len(time_col1), 1))

                # calculate time gaps matrix
                diffmat = np.subtract(time_col0, time_col1.T)
                # get tgaps both left and right
                # keep only negative ones
                masked_ar = np.ma.masked_where(
                    diffmat >= 0,
                    diffmat,
                )
                left_ar = np.max(masked_ar, axis=1)
                d_t_left = -left_ar.data[~left_ar.mask]

                # keep only positive ones
                masked_ar = np.ma.masked_where(diffmat <= 0, diffmat)
                right_ar = np.min(masked_ar, axis=1)
                d_t_right = right_ar.data[~right_ar.mask]

                d_t = np.concatenate([d_t_left.flatten(), d_t_right.flatten()])

        return d_t

    def run(self, data_slice, slice_point=None):
        # sort the data_slice in order of time.
        data_slice.sort(order=self.mjd_col)

        fom_dic = {}
        d_t_dic = {}
        for fltpair in self.fltpairs:
            d_t = self._get_d_t(data_slice, fltpair[0], fltpair[1])
            # calculate FoM
            # cut out at average
            # std below
            # FOM = Nv / std

            if fltpair[0] == fltpair[1]:
                bins = self.bins_same
            else:
                bins = self.bins_diff
                d_t_dic[fltpair] = d_t

            hist, _ = np.histogram(d_t, bins=bins)

            if np.any(hist):
                hist_clip = np.clip(hist, a_min=0, a_max=self.nv_clip[fltpair])
                fom = hist.sum() / hist_clip.std()
            else:
                fom = np.nan
            fom_dic[fltpair] = fom
            # print( fltpair, fom)

        return np.nansum(list(fom_dic.values()))



from .base_slicer import *
from .base_spatial_slicer import *
from .healpix_comcam_slicer import *
from .healpix_sdss_slicer import *
from .healpix_slicer import *
from .healpix_subset_slicer import *
from .hourglass_slicer import *
from .mo_slicer import *
from .movie_slicer import *
from .nd_slicer import *
from .one_d_slicer import *
from .time_interval_slicers import *
from .uni_slicer import *
from .user_points_slicer import *



"""
This file defines some test catalog and DBObject classes for use with unit tests.

To date (30 October 2014) testPhotometry.py and testCosmology.py import from this module
"""
__all__ = ("set_m5", "comoving_distance_integrand", "cosmological_omega")

import numpy

from rubin_sim.phot_utils import Sed, calc_sky_counts_per_pixel_for_m5


def set_m5(m5target, skysed, total_bandpass, hardware, phot_params, fwhm_eff=0.83):
    """
    Take an SED representing the sky and normalize it so that
    m5 (the magnitude at which an object is detected in this
    bandpass at 5-sigma) is set to some specified value.

    The 5-sigma limiting magnitude (m5) for an observation is
    determined by a combination of the telescope and camera parameters
    (such as diameter of the mirrors and the readnoise) together with the
    sky background. This method (set_m5) scales a provided sky background
    Sed so that an observation would have a target m5 value, for the
    provided hardware parameters. Using the resulting Sed in the
    'calcM5' method will return this target value for m5.

    Parameters
    ----------
    the : `Unknown`
        desired value of m5

    Parameters
    ----------
    skysed : `Unknown`
        is an instantiation of the Sed class representing
        sky emission

    Parameters
    ----------
    total_bandpass : `Unknown`
        is an instantiation of the Bandpass class
        representing the total throughput of the telescope (instrumentation
        plus atmosphere)

    Parameters
    ----------
    hardware : `Unknown`
        is an instantiation of the Bandpass class representing
        the throughput due solely to instrumentation.

    Parameters
    ----------
    phot_params : `Unknown`
        is an instantiation of the
        PhotometricParameters class that carries details about the
        photometric response of the telescope.

    Parameters
    ----------
    fwhm_eff : `Unknown`
        in arcseconds

    Parameters
    ----------
    returns : `Unknown`
        an instantiation of the Sed class that is the skysed renormalized
        so that m5 has the desired value.

    Note that the returned SED will be renormalized such that calling the method
    self.calcADU(hardwareBandpass) on it will yield the number of counts per square
    arcsecond in a given bandpass.
    """

    # This is based on the LSST SNR document (v1.2, May 2010)
    # www.astro.washington.edu/users/ivezic/Astr511/LSST_SNRdoc.pdf

    sky_counts_target = calc_sky_counts_per_pixel_for_m5(
        m5target, total_bandpass, fwhm_eff=fwhm_eff, phot_params=phot_params
    )

    sky_sed_out = Sed(wavelen=numpy.copy(skysed.wavelen), flambda=numpy.copy(skysed.flambda))

    sky_counts = (
        sky_sed_out.calc_adu(hardware, phot_params=phot_params)
        * phot_params.platescale
        * phot_params.platescale
    )
    sky_sed_out.multiply_flux_norm(sky_counts_target / sky_counts)

    return sky_sed_out


def cosmological_omega(redshift, h0, om0, ode0=None, og0=0.0, onu0=0.0, w0=-1.0, wa=0.0):
    """
    A method to compute the evolution of the Hubble and density parameters
    with redshift (as a baseline against which to test the cosmology unittest)

    Parameters
    ----------
    redshift : `Unknown`
        is the redshift at which the output is desired

    Parameters
    ----------
    h0 : `Unknown`
        is the Hubble parameter at the present epoch in km/s/Mpc

    Parameters
    ----------
    om0 : `Unknown`
        is the density parameter (fraction of critical) for matter at the
        present epoch

    Parameters
    ----------
    ode0 : `Unknown`
        is the density parameter for Dark Energy at the present epoch.
        If left as None, will be set to 1.0-om0-og0-onu0 (i.e. a flat universe)

    Parameters
    ----------
    og0 : `Unknown`
        is the density parameter for photons at the present epoch

    Parameters
    ----------
    onu0 : `Unknown`
        is the density parameter for neutrinos at the present epoch
        (assume massless neutrinos)

    Parameters
    ----------
    w0 : `Unknown`
        is a parameter for calculating the equation of state for Dark Energy
        w = w0 + wa * z/(1 + z)

    Parameters
    ----------
    wa : `Unknown`
        is the other parameter for calculating the equation of state for Dark
        Energy

    Returns
    -------
    Unknown: `Unknown`
        Hubble parameter at desired redshift (in km/s/Mpc)

    Returns
    -------
    Unknown: `Unknown`
        matter density Parameter at desired redshift

    Returns
    -------
    Unknown: `Unknown`
        Dark Energy density parameter at desired redshift

    Returns
    -------
    Unknown: `Unknown`
        photon density parameter at desired redshift

    Returns
    -------
    Unknown: `Unknown`
        neutrino density parameter at desired redshift

    Returns
    -------
    Unknown: `Unknown`
        curvature density parameter at desired redshift
    """

    if ode0 is None:
        ode0 = 1.0 - om0 - og0 - onu0

    ok0 = 1.0 - om0 - ode0 - og0 - onu0

    aa = 1.0 / (1.0 + redshift)
    omz = om0 * numpy.power(1.0 + redshift, 3)
    ogz = og0 * numpy.power(1.0 + redshift, 4)
    onuz = onu0 * numpy.power(1.0 + redshift, 4)
    okz = ok0 * numpy.power(1.0 + redshift, 2)
    odez = ode0 * numpy.exp(-3.0 * (numpy.log(aa) * (w0 + wa + 1.0) - wa * (aa - 1.0)))

    ototal = omz + ogz + onuz + odez + okz

    return (
        h0 * numpy.sqrt(ototal),
        omz / ototal,
        odez / ototal,
        ogz / ototal,
        onuz / ototal,
        okz / ototal,
    )


def comoving_distance_integrand(redshift, h0, om0, ode0, og0, onu0, w0, wa):
    """
    The integrand of comoving distance (as a baseline for cosmology unittest)

    Parameters
    ----------
    redshift : `Unknown`
        is the redshift at which to evaluate the integrand

    Parameters
    ----------
    h0 : `Unknown`
        is the Hubble parameter at the present epoch in km/s/Mpc

    Parameters
    ----------
    om0 : `Unknown`
        is the density parameter (fraction of critical) for matter at the
        present epoch

    Parameters
    ----------
    ode0 : `Unknown`
        is the density parameter for Dark Energy at the present epoch.

    Parameters
    ----------
    og0 : `Unknown`
        is the density parameter for photons at the present epoch

    Parameters
    ----------
    onu0 : `Unknown`
        is the density parameter for neutrinos at the present epoch
        (assume massless neutrinos)

    Parameters
    ----------
    w0 : `Unknown`
        is a parameter for calculating the equation of state for Dark Energy
        w = w0 + wa * z/(1 + z)

    Parameters
    ----------
    wa : `Unknown`
        is the other parameter for calculating the equation of state for Dark
        Energy

    Returns
    -------
    Unknown: `Unknown`
        1/(Hubble parameter at desired redshift in km/s/Mpc)

    """
    hh, mm, de, gg, nn, kk = cosmological_omega(
        redshift, h0, om0, ode0=ode0, og0=og0, onu0=onu0, w0=w0, wa=wa
    )
    return 1.0 / hh



# -*- coding: utf-8 -*-
"""
Created on Tue Sep 18 13:35:41 2018

@author: rstreet
"""
__all__ = ("calculate_lsst_field_visibility", "plot_visibility")

import copy
from sys import argv

# from astropy.visualization import astropy_mpl_style
# plt.style.use(astropy_mpl_style)
import astropy.units as u
import matplotlib.pylab as plt
import numpy as np
from astropy.coordinates import AltAz, EarthLocation, SkyCoord, get_sun
from astropy.time import Time, TimeDelta


def calculate_lsst_field_visibility(
    field_ra,
    field_dec,
    start_date,
    end_date,
    min_alt=30.0,
    dt_days=1.0,
    diagnostics=False,
    verbose=False,
):
    """Method to calculate the visibility of a given RA and Dec from LSST
    over the course of a year

    Adapted from an example in the Astropy docs.

    Inputs:
        :param float field_ra: Field RA in decimal degrees
        :param float field_dec: Field Dec in decimal degrees
        :param str start_date: Start date for calculations, UTC
        :param str end_date: End date for calculations, UTC
    """

    field = SkyCoord(field_ra, field_dec, frame="icrs", unit=(u.hourangle, u.deg))

    lsst = EarthLocation(
        lat=-30.239933333333333 * u.deg,
        lon=-70.7429638888889 * u.deg,
        height=2663.0 * u.m,
    )

    total_time_visible = 0.0
    t_start = Time(start_date + " 00:00:00")
    t_end = Time(end_date + " 00:00:00")
    cadence = 0.0007  # In days

    n_days = int((t_end - t_start).value)

    dates = np.array([t_start + TimeDelta(i, format="jd", scale=None) for i in range(0, n_days, 1)])

    target_alts = []
    hrs_visible_per_night = []
    hrs_per_night = []
    jds = []

    for d in dates:
        jds.append(d.jd)

        t = copy.copy(d)
        t.out_subfmt = "date"
        tstr = t.value

        intervals = np.arange(0.0, 1.0, cadence)

        dt = TimeDelta(intervals, format="jd", scale=None)

        ts = d + dt

        frame = AltAz(obstime=ts, location=lsst)

        altaz = field.transform_to(frame)

        alts = np.array((altaz.alt * u.deg).value)

        idx = np.where(alts > min_alt)[0]

        sun_altaz = get_sun(ts).transform_to(frame)

        sun_alts = np.array((sun_altaz.alt * u.deg).value)

        jdx = np.where(sun_alts < 12.0)[0]

        hrs_per_night.append(cadence * len(sun_alts[jdx]) * 24.0)

        idx = list(set(idx).intersection(set(jdx)))

        target_alts.append(alts[jdx].max())

        if len(idx) > 0:
            ts_vis = ts[idx]

            tvis = cadence * len(ts_vis)

            total_time_visible += tvis

            # target_alts.append(alts[idx].max())

            if verbose:
                print("Target visible from LSST for " + str(round(tvis * 24.0, 2)) + "hrs on " + tstr)

            hrs_visible_per_night.append((tvis * 24.0))

        else:
            # target_alts.append(-1e5)

            hrs_visible_per_night.append(0.0)

            if verbose:
                print("Target not visible from LSST on " + tstr)

    if diagnostics:
        plot_visibility(jds, target_alts, sun_alts, hrs_visible_per_night, min_alt)

    return total_time_visible, hrs_visible_per_night


def plot_visibility(ts, target_alts, sun_alts, hrs_visible_per_night, min_alt):
    """Function to plot a chart of the target and solar altitude above the
    horizon at the LSST site as a function of time"""

    ts = np.array(ts)
    target_alts = np.array(target_alts)

    (fig, ax1) = plt.subplots(figsize=(10, 10))

    plt.rcParams.update({"font.size": 18})
    plt.rc("xtick", labelsize=18)
    plt.rc("ytick", labelsize=18)
    plt.xticks(rotation=45.0)

    idx = np.where(target_alts > -1e5)
    ax1.plot((ts - 2450000)[idx], target_alts[idx], "b-", label="Target altitude")
    ax1.set_xlabel("JD")
    ax1.set_ylabel(r"Maximum altitude [$^{\circ}$]", color="b")
    ax1.xaxis.label.set_fontsize(18)
    ax1.yaxis.label.set_fontsize(18)
    for label in ax1.get_xticklabels():
        label.set_fontsize(18)
    for label in ax1.get_yticklabels():
        label.set_fontsize(18)

    t = [(ts - 2450000).min(), (ts - 2450000).max()]
    ax1.plot(t, [min_alt] * len(t), "g-.")

    ax1.grid(True)
    ax1.tick_params("y", colors="b")

    ax2 = ax1.twinx()
    ax2.plot(ts - 2450000, hrs_visible_per_night, "m--", label="Time target visible")
    ax2.set_ylabel("Hours per night", color="m")

    ax2.yaxis.label.set_fontsize(18)
    ax2.grid(False)
    ax2.tick_params("y", colors="m")

    fig.tight_layout()

    plt.legend()

    plt.savefig("target_visibility_from_lsst.png")

    plt.close()


if __name__ == "__main__":
    if len(argv) > 1:
        field_ra = argv[1]
        field_dec = argv[2]
        start_date = argv[3]
        end_date = argv[4]

    else:
        field_ra = input("Please enter the RA in sexigesimal format, J2000.0: ")
        field_dec = input("Please enter the Dec in sexigesimal format, J2000.0: ")
        start_date = input("Please enter the start date of the observing window, YYYY-MM-DD: ")
        end_date = input("Please enter the end date of the observing window, YYYY-MM-DD: ")

    (total_time_visible, hrs_per_night) = calculate_lsst_field_visibility(
        field_ra, field_dec, start_date, end_date, diagnostics=True
    )



"""MAF slicers to slice into time intervals based on MJD

Primarily intended for hourglass plots.
"""

# pylint: disable=too-many-arguments
__all__ = (
    "TimeIntervalSlicer",
    "BlockIntervalSlicer",
    "VisitIntervalSlicer",
    "SlicerNotSetupError",
)

from collections import defaultdict

# imports
from functools import wraps

import numpy as np
import pandas as pd

from .base_slicer import BaseSlicer


class SlicerNotSetupError(Exception):
    """Thrown when a slicer is not setup for the method called."""


# interface functions

# classes


class TimeIntervalSlicer(BaseSlicer):
    """Base for all time interval slicers.

    Slices in constant time intervals.

    Parameters
    ----------
    interval_seconds : `int`
        Duration of slice time intervals, in seconds
    mjd_column_name : `str`
        Name of column on which to slice, must be in units of days
    badval : `float`
        Value to use for bad values in slice
    verbose : `bool`
        Print extra information?
    """

    def __init__(
        self,
        interval_seconds=90,
        mjd_column_name="observationStartMJD",
        badval=np.NaN,
        verbose=False,
    ):
        super().__init__(verbose=verbose, badval=badval)
        self.interval_seconds = interval_seconds
        self.mjd_column_name = mjd_column_name
        self.columns_needed = [mjd_column_name]
        self.sim_idxs = defaultdict(list)  # pylint: disable=invalid-name

    def setup_slicer(self, sim_data, maps=None):
        visit_mjds = sim_data[self.mjd_column_name]
        start_mjd = np.floor(np.min(visit_mjds)).astype(int)
        end_mjd = np.ceil(np.max(visit_mjds)).astype(int)
        interval_days = self.interval_seconds / (24 * 60 * 60.0)

        mjd_bin_edges = np.arange(start_mjd, end_mjd + interval_days, interval_days)

        self.sim_idxs.update(
            pd.DataFrame(
                {
                    "visit_idx": np.arange(len(visit_mjds)),
                    "sid": np.digitize(visit_mjds, mjd_bin_edges),
                }
            )
            .groupby("sid")
            .agg(list)
            .to_dict()["visit_idx"]
        )

        mjds = np.arange(start_mjd, end_mjd, interval_days)
        self.slice_points["sid"] = np.arange(len(mjds))
        self.slice_points["mjd"] = mjds
        self.slice_points["duration"] = np.full_like(mjds, self.interval_seconds)
        self.nslice = len(mjds)
        self.shape = self.nslice
        self._run_maps(maps)

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):  # pylint: disable=invalid-name
            idxs = self.sim_idxs[islice]

            try:
                _ = idxs[0]
            except (TypeError, IndexError):
                idxs = [idxs]

            slice_points = {
                "sid": islice,
                "mjd": self.slice_points["mjd"][islice],
                "duration": self.slice_points["duration"][islice],
            }

            return {"idxs": idxs, "slice_point": slice_points}

        setattr(self, "_slice_sim_data", _slice_sim_data)

    def __eq__(self, other_slicer):
        """Evaluate if slicers are equivalent."""
        if not isinstance(other_slicer, self.__class__):
            return False

        for key in ["sid", "mjd", "duration"]:
            if not np.array_equal(other_slicer.slice_points[key], self.slice_points[key]):
                return False

        return True

    def _slice_sim_data(self, *args, **kwargs):
        raise SlicerNotSetupError()


class BlockIntervalSlicer(TimeIntervalSlicer):
    """Slices into intervals with common "note" values and no long gaps.

    Parameters
    ----------
    mjd_column_name : `str`
        Name of column on which to slice, must be in units of days
    duration_column_name : `str`
        Name of column with the duration of each visit (in seconds)
    note_column_name : `str`
        Name of column with the visit note.
    badval : `float`
        Value to use for bad values in slice
    verbose : `bool`
        Print extra information?
    """

    # Gap between visits in the same block, in hours
    gap_tolerance = 0.1

    def __init__(
        self,
        mjd_column_name="observationStartMJD",
        duration_column_name="visitTime",
        note_column_name="note",
        badval=np.NaN,
        verbose=False,
    ):
        super().__init__(verbose=verbose, badval=badval)
        self.mjd_column_name = mjd_column_name
        self.duration_column_name = duration_column_name
        self.note_column_name = note_column_name
        self.columns_needed = [
            mjd_column_name,
            duration_column_name,
            note_column_name,
        ]
        self.sim_idxs = defaultdict(list)  # pylint: disable=invalid-name

    def setup_slicer(self, sim_data, maps=None):
        visits = pd.DataFrame(
            sim_data,
            index=pd.Index(np.arange(len(sim_data[self.mjd_column_name])), name="visit_idx"),
        )
        visits.rename(
            columns={
                self.mjd_column_name: "mjd",
                self.duration_column_name: "duration",
                self.note_column_name: "note",
            },
            inplace=True,
        )
        # convert to hours
        visits.sort_values("mjd", inplace=True)
        visits["end_mjd"] = visits.mjd + visits.duration / (60 * 60 * 24.0)

        same_note = visits.note == visits.note.shift(-1)
        adjacent_times = visits.end_mjd + self.gap_tolerance / 24.0 > visits.mjd.shift(-1)
        visits["sid"] = np.logical_not(np.logical_and(same_note, adjacent_times)).cumsum().shift()
        visits["sid"].fillna(0, inplace=True)
        visits["sid"] = visits["sid"].astype(int)

        blocks = visits.groupby("sid").agg(
            mjd=pd.NamedAgg(column="mjd", aggfunc="min"),
            end_mjd=pd.NamedAgg(column="end_mjd", aggfunc="max"),
        )
        blocks["duration"] = (blocks.end_mjd - blocks.mjd) * 24 * 60 * 60

        self.nslice = len(blocks)
        self.shape = self.nslice
        self.sim_idxs.update(
            visits.reset_index()[["sid", "visit_idx"]].groupby("sid").agg(list).to_dict()["visit_idx"]
        )

        self.slice_points["sid"] = blocks.reset_index().sid.values
        self.slice_points["mjd"] = blocks.mjd.values
        self.slice_points["duration"] = blocks.duration.values
        self._run_maps(maps)

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):  # pylint: disable=invalid-name
            idxs = self.sim_idxs[islice]

            try:
                _ = idxs[0]
            except TypeError:
                idxs = [idxs]

            slice_points = {
                "sid": islice,
                "mjd": self.slice_points["mjd"][islice],
                "duration": self.slice_points["duration"][islice],
            }

            return {"idxs": idxs, "slice_point": slice_points}

        setattr(self, "_slice_sim_data", _slice_sim_data)


class VisitIntervalSlicer(TimeIntervalSlicer):
    """Slices into intervals each of which contain one visit

    Parameters
    ----------
    mjd_column_name : `str`
        Name of column on which to slice, must be in units of days
    duration_column_name : `str`
        Name of column with the duration of each visit (in seconds)
    badval : `float`
        Value to use for bad values in slice
    verbose : `bool`
        Print extra information?
    """

    def __init__(
        self,
        mjd_column_name="observationStartMJD",
        duration_column_name="visitTime",
        extra_column_names=tuple(),
        badval=np.NaN,
        verbose=False,
    ):
        super().__init__(verbose=verbose, badval=badval)
        self.mjd_column_name = mjd_column_name
        self.duration_column_name = duration_column_name
        self.extra_column_names = extra_column_names
        self.columns_needed = [mjd_column_name, duration_column_name]
        self.sim_idxs = None  # pylint: disable=invalid-name

    def setup_slicer(self, sim_data, maps=None):
        self.nslice = len(sim_data[self.mjd_column_name])
        self.shape = self.nslice

        self.sim_idxs = np.argsort(sim_data[self.mjd_column_name])
        self.slice_points["sid"] = np.arange(self.nslice)
        self.slice_points["mjd"] = sim_data[self.mjd_column_name]
        self.slice_points["duration"] = sim_data[self.duration_column_name]
        for column_name in self.extra_column_names:
            self.slice_points[column_name] = sim_data[column_name]
        self._run_maps(maps)

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):  # pylint: disable=invalid-name
            idxs = self.sim_idxs[islice]
            slice_points = {
                "sid": [idxs],
                "mjd": self.slice_points["mjd"][islice],
                "duration": self.slice_points["duration"][islice],
            }
            for column_name in self.extra_column_names:
                slice_points[column_name] = self.slice_points[column_name]

            return {"idxs": idxs, "slice_point": slice_points}

        setattr(self, "_slice_sim_data", _slice_sim_data)


# internal functions & classes
